import os
import time

def callscript(*args, **kwargs):
    cmd = kwargs['command']
    if 'argument' in kwargs.keys():
        cmd = cmd + " " + kwargs['argument']
    os.system(cmd)
    return 'Done!!! '

def callpython(message):
    print(message)


""" building and manipulating metal nanocluster models """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

from substrate import Substrate
from ase.constraints import FixAtoms

class Nanocluster(Substrate):
    """ construct and manipulate metallic nanoclusters """
    def __init__(self, params):
        super(Nanocluster, self).__init__(**params)

    def set_tags(self):
        graph = self.graph.copy()
        graph.vs['id'] = range(len(graph.vs))
        layer = 1
        tags = len(self.atoms)*[0]
        while len(graph.vs) > 0:
            ids = [s['id'] for s in graph.vs if graph.degree(s.index) != 12]
            ind = [s.index for s in graph.vs if graph.degree(s.index) != 12]
            for id_ in ids:
                tags[id_] = layer
            graph.delete_vertices(ind)
            layer += 1
        self.atoms.set_tags(tags)

    def set_constraints(self, constraints):
        if constraints is None:
            return
        if all('layers' not in c for c in constraints):
            super(Nanocluster, self).set_constraints(constraints)
            return
        for constraint in constraints:
            fixed = [i for i, tag in enumerate(self.atoms.get_tags())
                     if tag in constraint['layers']]
            self.atoms.set_constraint(FixAtoms(indices=fixed))



import time
import numpy as np
from random import random, seed
from ase.ga.data import PrepareDB
from ase.ga.data import DataConnection
from ase.ga.utilities import closest_distances_generator
from ase.ga.utilities import get_all_atom_types
from ase.ga.startgenerator import StartGenerator
from ase.ga.population import Population
from ase.ga.standard_comparators import InteratomicDistanceComparator
from ase.ga.cutandsplicepairing import CutAndSplicePairing
from ase.ga.offspring_creator import OperationSelector
from ase.ga.standardmutations import MirrorMutation
from ase.ga.standardmutations import RattleMutation
from ase.ga.standardmutations import PermutationMutation
from ase.ga.parallellocalrun import ParallelLocalRun
from ase.optimize import BFGS
from calculators import classmap
from ase.ga import get_parametrization
from ase.ga.utilities import get_atoms_connections, get_atoms_distribution
from ase.ga.utilities import get_angles_distribution
from ase.ga.utilities import get_rings, get_neighborlist


def combine_parameters(conf):
    # Get and combine selected parameters
    parameters = []
    gets = [get_atoms_connections(conf) + get_rings(conf) +
            get_angles_distribution(conf) + get_atoms_distribution(conf)]
    for get in gets:
        parameters += get
    return parameters


class GASimulation(object):
    """ The main class """

    # class wide objects
    da = None
    calc = None
    pairing = None
    mutations = None
    comparator = None
    phase = 0 # 1, 2
    parallel = False
    screening = False
    find_neighbors = None
    perform_parametrization = None

    def __init__(self, parameters):

        self.params = parameters
        self.ga_params = parameters['genetic algorithm']
        self.va_params = parameters['variate']
        self.calc_params = parameters['calculator']['parameters']

        self.check_state()
        print('Phase', self.phase)


    def check_state(self):
        # test whether the database exists
        if self.da is None:
            filename = str(self.ga_params['db file'])
            try:
                self.da = DataConnection(filename)
            except IOError as err:
                string = 'DB file ' + filename + ' not found'
                if string in err:
                    self.phase = 0
                    return
                else:
                    raise

        try:
            n_rel = len(self.da.get_all_relaxed_candidates())
            n_unr = self.da.get_number_of_unrelaxed_candidates()
        except:
            raise
        else:
            if n_unr > 0:
                self.phase = 1
            elif n_rel > 0:
                self.phase = 2
            else:
                self.phase = 0


    def create_initial_population(self):
        """ Constructs the substrate and the initial individuals """
        from ase import Atoms
        from ase.io import read
        from ase.constraints import FixAtoms
        from construct import create_structure

        self.check_state()
        assert self.phase == 0, 'population cannot be initialized twice'

# redundant to run-step.py

        # load / construct the substrate
        if self.params['substrate'] is None:
            self.substrate = Atoms()
        else:
            substr_params = self.params['substrate']
            if substr_params['input structure'] is not None:
                # load the structure from params
                substr = substr_params['input structure']
            elif substr_params['input file'] is not None:
                # read the structure from file
                substr = read(substr_params['input file'])
            else:
                # construct the structure
                substr = create_structure(**substr_params['construct'])

            # constraints
            if 'constrain all atoms' in substr_params:
                if substr_params['constrain all atoms']:
                    substr.set_constraint(FixAtoms(mask=len(substr) * [True]))
                elif substr_params['constraints']:
                    pass

            # magmoms
            if 'initial magnetic moments' in substr_params:
                magmoms = substr_params['initial magnetic moments']
                if type(magmoms) is list:
                    substr.set_initial_magnetic_moments(magmoms)
                else:
                    substr.set_initial_magnetic_moments(len(substr)*[magmoms])
            self.substrate = substr

# end of redundant section

        # construct the variate
        p0, v1, v2, v3 = self.get_variate_box()

        # define the closest distance between two atoms
        unique_atom_types = get_all_atom_types(self.substrate, 
            self.va_params['composition'])
        cd = closest_distances_generator(
            atom_numbers = unique_atom_types,
            ratio_of_covalent_radii = 
                self.ga_params['ratio of covalent radii']
        )

        # create the starting population
        sg = StartGenerator(slab=self.substrate,
                            atom_numbers=self.va_params['composition'],
                            closest_allowed_distances=cd,
                            box_to_place_in=[p0, [v1, v2, v3]])

        # generate the starting population
        self.starting_population = [sg.get_new_candidate() 
            for i in range(self.ga_params['population size'])]

        if 'initial magnetic moments' in self.va_params:
            for individual in self.starting_population:
                individual.set_initial_magnetic_moments(
                    self.substrate.get_initial_magnetic_moments().tolist() +
                    self.va_params['initial magnetic moments']
                )

        # initialize the database
        pdb = PrepareDB(
            db_file_name = str(self.ga_params['db file']),
            simulation_cell = self.substrate,
            stoichiometry = self.va_params['composition']
        )

        # save the starting population
        for a in self.starting_population:
            pdb.add_unrelaxed_candidate(a)


    def initialize_simulation(self):
        """ Initialize the different components of the GA """

        self.check_state()
        assert self.phase > 0, 'simulation can start only with some population'

        if 'number of procs' in self.ga_params.keys():
            if self.ga_params['number of procs'] > 1:
                self.parallel = True
                self.parallel_local_run = ParallelLocalRun(
                    data_connection = self.da,
                    tmp_folder = self.ga_params['temporary folder'],
                    n_simul = self.ga_params['number of procs'],
                    calc_script = self.ga_params['calc script']
                )
                print('Parallel mode turned on')

        atom_numbers_to_optimize = self.da.get_atom_numbers_to_optimize()
        n_to_optimize = len(atom_numbers_to_optimize)
        slab = self.da.get_slab()
        all_atom_types = get_all_atom_types(slab, atom_numbers_to_optimize)

        blmin = closest_distances_generator(
                all_atom_types,
                ratio_of_covalent_radii = 
                    self.ga_params['ratio of covalent radii']
        )

        self.comparator = InteratomicDistanceComparator(
            n_top = n_to_optimize,
            pair_cor_cum_diff = 0.015,
            pair_cor_max = 0.7,
            dE = 0.02,
            mic = False
        )

        self.pairing = CutAndSplicePairing(slab, n_to_optimize, blmin)

        if len(set(self.va_params['composition'])) != 1:
            ml = [1., 1., 1.]
            ms = [
                MirrorMutation(blmin, n_to_optimize),
                RattleMutation(blmin, n_to_optimize),
                PermutationMutation(n_to_optimize)
            ]
        else:
            ml = [1., 1.]
            ms = [
                MirrorMutation(blmin, n_to_optimize),
                RattleMutation(blmin, n_to_optimize)
            ]
        self.mutations = OperationSelector(ml, ms)

        if 'rng seed' in list(self.ga_params.keys()):
            seed(self.ga_params['rng seed'])

        if 'screening' in list(self.ga_params.keys()):
            if self.ga_params['screening'] is not None:
                self.screening = True
                self.find_neighbors = get_neighborlist
                self.perform_parametrization = combine_parameters


    def relax_structures(self):
        """ Relax all unrelaxed structures (e.g. the starting population) """

        self.check_state()
        assert self.phase > 0, 'no structures to optimize'

        self.create_calculator()
        while self.da.get_number_of_unrelaxed_candidates() > 0:
            a = self.da.get_an_unrelaxed_candidate()
            self.relax_structure(a)

        if self.parallel:
            while self.parallel_local_run.get_number_of_jobs_running() > 0:
                time.sleep(5.)


    def create_population(self):
        if self.screening:
            self.create_population_with_screening()
        else:
            self.create_population_without_screening()


    def create_population_without_screening(self):
        """ Create new population """

        self.check_state()
        assert self.phase > 1, 'all structures must be relaxed first'
        assert all((self.comparator, self.pairing, self.mutations))

        self.create_calculator()

        population = Population(
            data_connection = self.da,
            population_size = self.ga_params['population size'],
            comparator = self.comparator
        )

        """ test new candidates """
        for i in range(self.ga_params['individuals to test']):
            print('Starting configuration number {0}'.format(i))
            a1, a2 = population.get_two_candidates()
            a3, desc = self.pairing.get_new_individual([a1, a2])
            if a3 is None:
                continue
            self.da.add_unrelaxed_candidate(a3, description=desc)

            # Check if we want to do a mutation
            if random() < self.ga_params['mutation probability']:
                a3_mut, desc = self.mutations.get_new_individual([a3])
                if a3_mut is not None:
                    self.da.add_unrelaxed_step(a3_mut, desc)
                    a3 = a3_mut

            # Relax the new candidate
            self.relax_structure(a3)
            population.update()

        if self.parallel:
            while self.parallel_local_run.get_number_of_jobs_running() > 0:
                time.sleep(5.)


    def create_population_with_screening(self):
        """ Create new population """
        self.check_state()
        assert self.phase > 1, 'all structures must be relaxed first'
        assert all((self.comparator, self.pairing, self.mutations))

        self.create_calculator()

        population = Population(
            data_connection = self.da,
            population_size = self.ga_params['population size'],
            comparator = self.comparator
        )

        # create the regression expression for estimating the energy
        all_trajs = self.da.get_all_relaxed_candidates()
        sampled_points = []
        sampled_energies = []
        for conf in all_trajs:
            no_of_conn = list(get_parametrization(conf))
            if no_of_conn not in sampled_points:
                sampled_points.append(no_of_conn)
                sampled_energies.append(conf.get_potential_energy())

        sampled_points = np.array(sampled_points)
        sampled_energies = np.array(sampled_energies)

        if len(sampled_points) > 0 and len(sampled_energies) >= len(sampled_points[0]):
            weights = np.linalg.lstsq(sampled_points, sampled_energies)[0]
        else:
            weights = None

        """ Test new candidates """
        for i in range(self.ga_params['individuals to test']):
            print('Starting configuration number {0}'.format(i))
            a1, a2 = population.get_two_candidates()

            # Select the "worst" parent energy to which the child will be compared
            ce_a1 = self.da.get_atoms(a1.info['relax_id']).get_potential_energy()
            ce_a2 = self.da.get_atoms(a2.info['relax_id']).get_potential_energy()
            comparison_energy = min(ce_a1, ce_a2)

            a3, desc = self.pairing.get_new_individual([a1, a2])
            if a3 is None:
                continue
            if self.should_we_skip(a3, comparison_energy, weights):
                continue
            self.da.add_unrelaxed_candidate(a3, description=desc)

            if random() < self.ga_params['mutation probability']:
                a3_mut, desc_mut = self.mutations.get_new_individual([a3])
                if (a3_mut is not None and
                    not self.should_we_skip(a3_mut, comparison_energy, weights)):
                    self.da.add_unrelaxed_step(a3_mut, desc_mut)
                    a3 = a3_mut

            # Relax the new candidate
            self.relax_structure(a3)
            population.update()

        if self.parallel:
            while self.parallel_local_run.get_number_of_jobs_running() > 0:
                time.sleep(5.)


    def should_we_skip(self, conf, comparison_energy, weights):
        parameters = combine_parameters(conf)
        st = self.ga_params['screening']['skip threshold']
        sp = self.ga_params['screening']['skip probability']
        # Return if weights not defined (too few completed
        # calculated structures to make a good fit)
        if weights is None:
            return False
        regression_energy = sum(p * q for p, q in zip(weights, parameters))
        # Skip with sp likelihood if energy appears to go up st eV or more
        if (regression_energy - comparison_energy) > st and random() < sp:
            return True
        else:
            return False


    def get_population(self):
        from ase.io import write
        write('all_candidates.traj', self.da.get_all_relaxed_candidates())


    def get_variate_box(self):
        """ Define the volume in which the variate structure is optimized
        the volume is defined by a corner position (p0)
        and three spanning vectors (v1, v2, v3) """

        par = self.va_params['box']
        if par['type'] == 'ontop':
            pos = self.substrate.get_positions()
            cell = self.substrate.get_cell()
            p0 = np.array([0., 0., max(pos[:, 2]) + par['origin']])
            v1 = cell[0, :] * par['v1 factor']
            v2 = cell[1, :] * par['v2 factor']
            v3 = cell[2, :]
            v3[2] = par['v3 value']
        elif par['type'] == 'centered':
            a = par['outer']
            b = par['inner']
            self.substrate.set_cell([a, a, a])
            cell = self.substrate.get_cell()
            p0 = np.array([(a-b)/2, (a-b)/2, (a-b)/2]) 
            v1 = cell[0, :]*b/a
            v2 = cell[1, :]*b/a
            v3 = cell[2, :]*b/a
        else:
            p0, v1, v2, v3 = ()
        return p0, v1, v2, v3


    def open_db(self):
        self.da = DataConnection(str(self.ga_params['db file']))


    def show_initial_population(self):
        from ase.visualize import view
        view(self.starting_population)


    def create_calculator(self):
        if self.calc is None:
            calc_name = self.params['calculator']['name']
            self.calc = classmap[calc_name](**self.calc_params)


    def relax_structure(self, a):
        import warnings
        if self.parallel:
            self.parallel_local_run.relax(a)
        else:
            a.set_calculator(self.calc)
            print('Relaxing starting candidate {0}'.format(a.info['confid']))
            try:
                if self.params['calculator']['optimize']:
                    self.calc.calculate(a)
                else:
                    dyn = BFGS(a, trajectory=None, logfile=None)
                    dyn.run(
                        fmax = abs(self.calc_params['ediffg']),
                        steps = self.calc_params['nsw']
                    )
                energy = a.get_potential_energy()
                if hasattr(self.calc, 'converged'):
                    assert self.calc.converged
                assert energy is not None
                a.info['key_value_pairs']['raw_score'] = - energy
            except(RuntimeError, AssertionError):
                self.da.c.update(a.info['confid'], relaxed=-1)
                message = 'Candidate {0} will be skipped'.format(a.info['confid'])
                warnings.warn(message)
            else:
                self.da.add_relaxed_step(a, find_neighbors=self.find_neighbors, 
                    perform_parametrization=self.perform_parametrization)





""" A collection of functions to compute surface Pourbaix diagrams """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

from data import Atoms
from data import atoms2dict
from data import single_entry
from calculators import classmap

def structure_relaxation(structure, test=False):
    """ perform structure relaxation """

    structure = single_entry(structure)
    if test:
        return {
            'atoms': structure['atoms'],
            'properties': {
               'energy': 0.0,
               'forces': None,
               'dipole': None
            },
            'parameters': structure['parameters']
        }

    atoms = Atoms.from_dict(structure['atoms'])
    params = structure['parameters']
    calc_params = params['calculator']['parameters']
    calc_name = str(params['calculator']['name'])
    if 'ropt' in calc_params:
        ropt = calc_params['ropt']
        if not isinstance(ropt, list):
            calc_params['ropt'] = len(set(atoms.get_atomic_numbers()))*[ropt]
    calc = classmap[calc_name](**calc_params)
    atoms.set_calculator(calc)
    calc.calculate(atoms)
    assert calc.converged

    adict = atoms2dict(atoms)
    energy = adict.pop('energy')
    forces = adict.pop('forces')
    dipole = adict.pop('dipole')
    return {
        'atoms': adict,
        'properties': {
           'energy': energy,
           'forces': forces,
           'dipole': dipole
        },
        'parameters': params
    }

def free_energy(entry, ads_props):
    """ approximated free energy using average values for ZPE and entropy """
    entry = single_entry(entry)
    zpe = 0.0
    vet = 0.0
    for ads in entry['parameters']['adsorbates']:
        atoms = Atoms.from_dict(ads['adsorbate'])
        adskey = atoms.get_chemical_formula(mode='hill').lower()
        zpe += ads_props[adskey]['zero-point energy']
        vet += ads_props[adskey]['vibrational entropy term']
    entry['properties']['free energy'] = entry['properties']['energy'] + zpe + vet
    entry['properties']['zero-point energy'] = zpe
    entry['properties']['total entropy term'] = vet
    # assuming energy minimum
    entry['properties']['energy minimum'] = True
    return entry

def substrate_entry(entries):
    """ detect and select the substrate entry """
    substr_entry = [e for e in entries if len(e['parameters']['adsorbates'])==0]
    try:
        entry = single_entry(substr_entry)
    except AssertionError:
        if len(substr_entry) == 0:
            print('substrate entry not found')
        if len(substr_entry) > 1:
            print('several substrate entries found')
        raise

    entry['properties']['free energy'] = entry['properties']['energy']
    entry['properties']['zero-point energy'] = 0.000
    entry['properties']['total entropy term'] = 0.000
    entry['properties']['energy minimum'] = True
    return entry

def formation_energy(entry, substrate, gas_entries):
    """ calculate formation energy of the complex from elemental species """
    from ueffumax import reaction_energies
    from data import Atoms

    entry = single_entry(entry)
    props = {}
    props['complex'] = entry['properties']
    props['substrate'] = substrate['properties']
    props.update(gas_entries)

    atoms = Atoms.from_dict(entry['atoms'])
    nhydr = len([s for s in atoms.get_chemical_symbols() if s == 'H'])
    noxyg = len([s for s in atoms.get_chemical_symbols() if s == 'O'])
    reactions = [{
      'equation': [
        {'species': 'substrate', 'coefficient': -1.},
        {'species': 'h2', 'coefficient': -nhydr/2.},
        {'species': 'o2', 'coefficient': -noxyg/2.},
        {'species': 'complex', 'coefficient': 1.}
      ]
    }]
    reactions = reaction_energies(props, reactions)
    entry['properties']['reactions'] = reactions
    keylist = ['energy', 'zero-point energy', 'zpe-corrected energy',
               'total entropy term', 'free energy']
    for key in keylist:
        entry['properties'][key] = reactions[0][key]
    return entry

def pourbaix_diagram(ion_entries, comp_entries):
    """ construct a surface Pourbaix diagram """
    from pymatgen.entries.computed_entries import ComputedEntry
    from pymatgen.analysis.pourbaix_diagram import PourbaixEntry
    from pymatgen.analysis.pourbaix_diagram import PourbaixDiagram
    from pymatgen.analysis.pourbaix_diagram import PourbaixPlotter

    if all(isinstance(e, PourbaixEntry) for e in ion_entries):
        entries = ion_entries
    else:
        entries = [PourbaixEntry.from_dict(edict) for edict in ion_entries]
    for entry in comp_entries:
        atoms = Atoms.from_dict(entry['atoms'])
        composition = atoms.get_chemical_formula(mode='hill')
        energy = entry['properties']['free energy']
        entries.append(PourbaixEntry(ComputedEntry(composition, energy)))
    pbx = PourbaixDiagram(entries)
    plotter = PourbaixPlotter(pbx)
    plot = plotter.get_pourbaix_plot()
    plot.savefig('fig.png')



""" helper functions for manipulating surfaces """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

import numpy as np

def modify_surface_atom(slab, symbol, offset=None):
    """ modify surface atom by specified offset """

    info = slab.info.get('adsorbate_info', {})
    spos = np.array([0.0, 0.0])
    if offset is not None:
        spos += np.asarray(offset, float)

    if 'cell' in info:
        cell = info['cell']
    else:
        cell = slab.get_cell()[:2, :2]

    pos = np.dot(spos, cell)

    # get the z-coordinate:
    if 'top layer atom index' in info:
        atom = info['top layer atom index']
    else:
        atom = slab.positions[:, 2].argmax()
        if 'adsorbate_info' not in slab.info:
            slab.info['adsorbate_info'] = {}
        slab.info['adsorbate_info']['top layer atom index'] = atom

    zcoord = slab.positions[atom, 2]
    distances = slab.positions - np.array([pos[0], pos[1], zcoord])
    indices = np.argwhere(np.isclose(np.linalg.norm(distances, axis=1), 0))
    for index in list(indices.flatten()):
        slab[index].set('symbol', symbol)



import node_1001_1105_parsl
#import node_501_592_parsl


if __name__ == '__main__':
    
    core_number = [80, 70 , 60, 50, 40, 30, 20, 10, 8, 4, 2, 1]
    
    for core in core_number:
        node_1001_1105_parsl.run_model(core)
    



""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'




import os
import os.path
from copy import deepcopy
from collections import OrderedDict
import shutil
import argparse
import json
import xmlschema
import yaml
import jsonschema
from jsonschema import validate
ROOT_DIR = os.path.dirname(os.path.abspath(__file__)) 

class BasewfGenes():
    """ Perform array initialization and data-flow analysis """
    def __init__(self, blueargs):
        self.args_workflowconfig = blueargs['workflowconfig']
        self.args_inputpath = blueargs['inputpath']
        self.args_wms = blueargs['wms']
        self.previous_stamp = os.stat(self.args_workflowconfig).st_mtime
        # List declaration
        schema_path = os.path.join(ROOT_DIR, 'wconfig_schema.json')
        with open(self.args_workflowconfig) as input_stream:
            schema = json.load(open(schema_path))
            config_name_split = self.args_workflowconfig.split('.')
            num_dot = len(config_name_split) - 1
            if config_name_split[num_dot] == 'yaml' or config_name_split[num_dot] == 'yml':
                self.interface_dict = yaml.load(input_stream, Loader=yaml.Loader)
                validate(self.interface_dict , schema=schema)
            elif config_name_split[num_dot] == 'json':
                self.interface_dict = json.load(input_stream)
                validate(self.interface_dict , schema=schema)
              
            for k, v in self.interface_dict.items():
                if k == 'nodes':
                    routine_interface = deepcopy(v)
            
            if 'metadata' in self.interface_dict: 
                self.metadata = self.interface_dict['metadata']
            else: 
                self.metadata = {'type of generation': 'By wfGenes'}
            self.routine_number = len(self.interface_dict['nodes'])
            self.subroutine_number = [0 for y in range(self.routine_number)]
            self.routine_name = [0 for y in range(self.routine_number)]
            self.routine_dir = [0 for y in range(self.routine_number)]
            self.routine_gdep = [0 for y in range(self.routine_number)]
            for i in range(self.routine_number):
                self.subroutine_number[i] = len(routine_interface[i]['tasks'])
            self.func = [['null' for x in range(self.subroutine_number[y])]
                        for y in range(self.routine_number)]
            self.func_nodup = [['null' for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.func_global_nodup = [['null' for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.func_no_locdep = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]        
            self.func_depid = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]
            self.func_localdepid = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]                    
            self.inputs = [['null' for x in range(self.subroutine_number[y])]
                        for y in range(self.routine_number)]

            self.inputs_py = [[0 for x in range(self.subroutine_number[y])]
                        for y in range(self.routine_number)]

            self.inputs_nodup = [['null' for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.inputs_no_locdep = [['null' for x in range(self.subroutine_number[y])]
                                        for y in range(self.routine_number)]
            self.inputs_locname = [['null' for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.inputs_links = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]
            
            self.inputs_global_nodup = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]

            self.outputs = [['null' for x in range(self.subroutine_number[y])]
                            for y in range(self.routine_number)]
            self.outputs_py = [['null' for x in range(self.subroutine_number[y])]
                            for y in range(self.routine_number)]

            self.outputs_nodup = [['null' for x in range(self.subroutine_number[y])]
                                        for y in range(self.routine_number)]
            self.outputs_no_locdep = [['null' for x in range(self.subroutine_number[y])]
                                        for y in range(self.routine_number)]
            self.outputs_locname = [['null' for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.foreach_output = [[['null' for k in range(3)] for x in range(self.subroutine_number[y])]
                                    for y in range(self.routine_number)]
            self.kwargs = [['null' for x in range(self.subroutine_number[y])]
                    for y in range(self.routine_number)]
            self.kwargs_nodup = [['null' for x in range(self.subroutine_number[y])]
                                for y in range(self.routine_number)]

            self.resource = [['null' for x in range(self.subroutine_number[y])]
                    for y in range(self.routine_number)]                    
            
            self.firstround_scheduling = True                  

                               

            # Start of yaml parser
            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    if 'func' in  routine_interface[i]['tasks'][j]:
                        self.func[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['func'])
                    elif 'script' in  routine_interface[i]['tasks'][j]:
                        self.func[i][j] = 'ScriptTask'
            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    self.func_nodup[i][j] = deepcopy(self.func[i][j])
                    self.func_global_nodup[i][j]= deepcopy(self.func[i][j])
                    self.func_no_locdep[i][j]= deepcopy(self.func[i][j])
                    #self.func_depid[i][j]= deepcopy(routine_interface[i]['tasks'][j]['func'][1])  
                self.func_nodup[i][:] = self.extractlist(self.func_nodup[i][:])
                self.func_global_nodup[i][:] = self.extractlist(self.func_global_nodup[i][:])
                self.func_depid[i][:] = self.extractlist(self.func_depid[i][:])
                self.func_localdepid[i][:] = self.extractlist(self.func_localdepid[i][:])
                self.func_no_locdep[i][:] = self.extractlist(self.func_no_locdep[i][:])

            for i in range(self.routine_number):
                self.func_nodup[i][:] = self.duplicate_finder(
                    self.func_nodup[i][:])

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    if 'inputs' in routine_interface[i]['tasks'][j]:
                        self.inputs[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['inputs'])
                        self.inputs_py[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['inputs'])
                    elif 'inputs' not in routine_interface[i]['tasks'][j]:       
                        self.inputs[i][j] = ['NULL']
                        self.inputs_py[i][j] = ['NULL']
            self.pylist(self.inputs_py)      

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    self.inputs_nodup[i][j] = deepcopy( self.inputs_py[i][j])
                    self.inputs_global_nodup[i][j] =  deepcopy( self.inputs_py[i][j]) 
                    self.inputs_no_locdep[i][j] =  deepcopy( self.inputs_py[i][j])
                    self.inputs_links[i][j] =  deepcopy( self.inputs_py[i][j])
                    self.inputs_locname[i][j] =  deepcopy( self.inputs_py[i][j])
                    
            for i in range(self.routine_number):
                self.inputs_nodup[i][:] = self.duplicate_finder(
                    self.inputs_nodup[i][:])

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    if 'outputs' in routine_interface[i]['tasks'][j]:
                        self.outputs[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['outputs'])
                        self.outputs_py[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['outputs'])
                    elif 'outputs' not in routine_interface[i]['tasks'][j]:
                        self.outputs[i][j] = ['NULL']
                        self.outputs_py[i][j] = ['NULL']


            self.pylist(self.outputs_py)
            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    self.outputs_nodup[i][j] = deepcopy(self.outputs_py[i][j])
                    self.outputs_locname[i][j] = deepcopy(self.outputs_py[i][j])         
            for i in range(self.routine_number):
                self.outputs_nodup[i][:] = self.duplicate_finder(
                    self.outputs_nodup[i][:])      

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    for k in range(len(self.inputs_locname[i][j])):
                        self.inputs_locname[i][j][k] = self.inputs_locname[i][j][k].rsplit('_id', 1)[
                            0]

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    for k in range(len(self.outputs_locname[i][j])):
                        self.outputs_locname[i][j][k] = self.outputs_locname[i][j][k].rsplit('_id', 1)[
                            0]

            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    if 'kwargs' in routine_interface[i]['tasks'][j]:
                        self.kwargs[i][j] = deepcopy(routine_interface[i]['tasks'][j]['kwargs'])
                        self.kwargs_nodup[i][j] = deepcopy(
                            routine_interface[i]['tasks'][j]['kwargs'])

            
            for i in range(self.routine_number):
                for j in range(len(routine_interface[i]['tasks'])):
                    if 'resource' in routine_interface[i]['tasks'][j]:
                        self.resource[i][j] = deepcopy(routine_interface[i]['tasks'][j]['resource'])


          

            for i in range(self.routine_number):
                self.routine_name[i] = routine_interface[i]['name']
                self.routine_dir[i] = routine_interface[i]['name']
            self.pyvar(self.routine_dir)    


            for i in range(self.routine_number):
                self.inputs_no_locdep[i], self.func_no_locdep[i][:] = self.dependency_check(
                    self.inputs_no_locdep[i][:], self.outputs_py[i][:], i, self.func_localdepid[i][:])
                self.inputs_links[i], self.func_depid[i][:]  = self.dependency_check(
                    self.inputs_links[i][:], self.outputs_py[i][:], i, self.func_depid[i][:])
                        
            # Global dependency analysis
            for i in range(self.routine_number):
                for j in range(self.routine_number):
                    if i != j:
                        self.inputs_links[i][:], self.func_depid[i][:], self.routine_gdep[i] = self.link_check(
                            self.inputs_links[i][:], self.inputs_no_locdep[i][:], self.outputs_py[j][:], j, self.func_depid[i][:], self.routine_gdep[i])
                        
            for i in range(self.routine_number):
                for j in range(i+1, self.routine_number):
                    # Functions and input name dunplication is not allowed- Should be validate with schema            
                    self.func_global_nodup[i][:] = self.global_duplication_check(self.func_global_nodup[i][:], self.func_global_nodup[j][:])
                    self.inputs_global_nodup[i][:] = self.global_duplication_check(self.inputs_global_nodup[i][:], self.inputs_global_nodup[j][:])
                    
            self.foreach_dependency()
            self.link_generation()


            self.interface_dict['workflow_name'] = self.interface_dict['workflow_name'].replace(" ","_")
            wf_name_split = self.interface_dict['workflow_name'].split('.')
            last_dot_index = len(wf_name_split) - 1
            if last_dot_index > 0:
                self.interface_dict['workflow_name'] = wf_name_split[0] 
            self.workflow_path = os.path.join(
                os.path.dirname(os.path.abspath(self.args_workflowconfig)),
                'wfGenes_output',
                self.interface_dict['workflow_name'])

            if not os.path.exists(self.workflow_path):
                os.makedirs(self.workflow_path)

    def extractlist(self, lst):
        """ Return list of the list """
        return [[el] for el in lst]
            
    def pylist(self, input_py):
        """ Replace ' ' with '_' to be used for python wrappers """
        map_list = [[" ", "_"], ["*", "star"]]
        for couple in map_list:
            for i in range(len(input_py)):
                for j in range(len(input_py[i])):
                    for k in range(len(input_py[i][j])):
                        input_py[i][j][k] = input_py[i][j][k].replace(couple[0], couple[1])
    def pyvar(self, input_py):
        """ Replace ' ' with '_' to be used for python wrappers """
        map_list = [[" ", "_"], ["*", "star"]]
        for couple in map_list:
            for i in range(len(input_py)):
                input_py[i]= input_py[i].replace(couple[0], couple[1])

    def foreach_dependency(self):  
        """ Check dependency for split variable in foreach task """
        for i in range(self.routine_number):
            for j in range(self.subroutine_number[i]):                                    
                if len(self.func[i][j]) > 2:
                    if self.func[i][j][2] == 'FOREACH':
                        for k in range(self.routine_number):
                            if k != i:
                                for m in range(self.subroutine_number[k]):
                                    if self.func[i][j][3] in self.outputs[k][m]:
                                        # Iterable variable for dpendent foreach
                                        self.foreach_output[i][j] =  [self.func[k][m][1] , k , m]
                                            
    def duplicate_finder(self, target_list):
        """ Detect duplicates in multidimensional list  """
        for y in range(len(target_list)):
            for x in range(len(target_list[y])):
                for i in range(y, len(target_list) - 1):
                    for j in range(len(target_list[i + 1])):
                        if target_list[y][x] != target_list[i + 1][j]:
                            pass
                        else:
                            target_list[i + 1][j] = 'duplicate'
        return target_list

    def dependency_check(self, input_list, output_list, id, func):
        """ Check dependency between different subroutine in one routine """
        for y in range(len(input_list) - 1, -1, -1):
            for x in range(len(input_list[y])):
                for j in range(len(output_list) - 1):
                    for k in range(len(output_list[j])):
                        if input_list[y][x] != output_list[j][k]:
                            pass
                        else:
                            input_list[y][x] = ['inner_dependent', str(j) , str(k)]
                            func[y].append([id , j])
        return input_list, func

    def link_check(self, input_link, input_list, output_list, id, func, routine):
        """ Check dependency between different routines"""
        for y in range(len(input_list) - 1, -1, -1):
            for x in range(len(input_list[y])):
                for j in range(len(output_list) - 1, -1, -1):
                    for k in range(len(output_list[j])):
                        if input_list[y][x] != output_list[j][k]:
                            pass
                        else:
                            input_link[y][x] = [str(id), str(j) , str(k)]
                            func[y].append([id, j])
                            routine = True
        return input_link, func, routine

    def global_duplication_check(self, input_list, output_list):
        """ Check for duplication between diffrent nodes """
        for y in range(len(input_list) - 1, -1, -1):
            for x in range(len(input_list[y])):
                for i in range(len(output_list)):
                    for j in range(len(output_list[i])):
                        if input_list[y][x] != output_list[i][j]:
                            pass
                        else:
                            output_list[i][j] = 'global_duplicate'
        return input_list    

    def wfgenes_scheduler(self, done_func):
        """ Update dependency list to choose ready function(s)"""
        self.func_waiting = False
        for routine in self.func_depid :
            for subroutine in routine:
                for routine_name in subroutine:
                    if routine_name != 'null' or self.firstround_scheduling == True:
                        self.func_waiting= True
                        if self.firstround_scheduling == True:
                            self.firstround_scheduling = False
                        break
        for func in done_func:
            for i in range(self.routine_number):
                for j in range(self.subroutine_number[i]):
                    while func in self.func_depid[i][j]: 
                        self.func_depid[i][j].remove(func) 
                        # Remove implemented function from func_depid

    
    def wconfig_checker(self, args= 'none'):
        """ Check if wconfig is verified by user to proceed to code generation """
        modeling = 'default' #
        input_stream = open(self.args_workflowconfig, 'r')
        config_name_split = self.args_workflowconfig.split('.')
        if config_name_split[1] == 'yaml':
            interface_dict = yaml.load(input_stream, Loader=yaml.Loader)
        elif config_name_split[1] == 'json':
            interface_dict = json.load(input_stream)

        wf_name_split = self.interface_dict['workflow_name'].split('.')
        last_dot_index = len(wf_name_split) - 1
        
        if last_dot_index > 0:
            if wf_name_split[last_dot_index] == 'active':
                modeling = 'active'
                self.interface_dict['workflow_name'] = wf_name_split[0]    

            elif wf_name_split[last_dot_index] == 'done':
                modeling = 'done'
                self.interface_dict['workflow_name'] = wf_name_split[0]    
            else :
                modeling = 'default'
                self.interface_dict['workflow_name'] = wf_name_split[0]

        wf_path = os.path.join(os.path.dirname(os.path.abspath(self.args_workflowconfig)),
                'wfGenes_output',
                self.interface_dict['workflow_name'])

        if args == 'clean':
            if os.path.exists(wf_path):
                shutil.rmtree(wf_path)

        
        stamp = os.stat(self.args_workflowconfig).st_mtime
        if stamp != self.previous_stamp:
            self.previous_stamp = stamp
            wconfig_stat = 'modified'
        else:
            wconfig_stat = 'unchanged'    

        return modeling, wf_path, wconfig_stat

    def link_generation(self):
        self.links = {}
        for M in range(self.routine_number):
            self.links[str(M)] = []
            for i in range(self.routine_number):
                for j in range(self.subroutine_number[i]):
                    for k in range(len(self.inputs_links[i][j])):
                        if isinstance(
                                self.inputs_links[i][j][k],
                                list) and self.inputs_links[i][j][k][0] == str(M):
                            self.links[str(M)].append(i)             
            self.links[str(M)] = list(OrderedDict.fromkeys(self.links[str(M)]))

               
                                
        


""" a simple interface to query fireworks and workflows """


def get_launch(lpad, query):
    """ return a single firework from a query and its most recent launch """
    fw_list = list(lpad.fireworks.find(query))
    if len(fw_list) == 0:
        return None, None
    assert len(fw_list) == 1
    fw = fw_list[0]
    if len(fw['launches']) > 0:
        launch = lpad.launches.find_one({'launch_id': fw['launches'][-1]})
    else:
        launch = None
    return fw, launch


def db_select(lpad, filters={}, ids=[], selects=[]):
    """ apply workflow query filters and then select updates from launch """

    assert isinstance(filters, dict) or ids
    if not ids:
        wfq = filters.get('workflows', {})
        fwq = filters.get('fireworks', {})
        projection = {'nodes': True}
        wfns = [i['nodes'] for i in lpad.workflows.find(wfq, projection)]
        if fwq:
            projection = {'fw_id': True}
            fws = [i['fw_id'] for i in lpad.fireworks.find(fwq, projection)]
            wfnsf = [wfn[0] for wfn in wfns if any(i in wfn for i in fws)]
        else:
            wfnsf = [wfn[0] for wfn in wfns]
    else:
        wfnsf = ids

    result = []
    for wf_id in wfnsf:
        wf = lpad.workflows.find_one({'nodes': wf_id})
        wf_data = {}
        wf_data['name'] = wf['name']
        wf_data['metadata'] = wf['metadata']
        wf_data['state'] = wf['state']
        wf_data['fws'] = []
        for select in selects:
            query = {'name': select['fw_name'], 'fw_id': {'$in': wf['nodes']}}
            fw, launch = get_launch(lpad, query)
            fw_data = {}
            if fw:
                fw_data['name'] = select['fw_name']
                fw_data['id'] = fw['fw_id']
                fw_data['updated_on'] = fw['updated_on']
                fw_data['created_on'] = fw['created_on']
                fw_data['state'] = fw['state']
                fw_data['parents'] = wf['parent_links'].get(str(fw['fw_id']))
                if select.get('add fw_spec', False):
                    fw_data['spec'] = fw['spec']
                if launch and launch.get('action'):
                    outputs = select.get('fw_updates')
                    updates = launch['action'].get('update_spec')
                    if isinstance(outputs, list):
                        fw_data['updates'] = {o: updates[o] for o in outputs}
                    elif outputs:
                        fw_data['updates'] = updates
                    else:
                        fw_data['updates'] = None
                else:
                    fw_data['updates'] = None
                if launch:
                    fw_data['launch_dir'] = launch.get('launch_dir')
            wf_data['fws'].append(fw_data)
        result.append(wf_data)
    return result



from wfgenes.core.initial_wfgenes import BasewfGenes
import os
import os.path
from copy import deepcopy
from collections import OrderedDict
import argparse
import json
import xmlschema
import yaml


class SimstackwfGenes(BasewfGenes):


    def __init__(self, blueargs):
        BasewfGenes.__init__(self, blueargs)
        self.wano_genrator()
        self.workflow_generator()
        self.pack_wano()


    def wano_genrator(self):
        for i in range(self.routine_number):
            self.simstack_path = os.path.join(self.workflow_path, 'SimStack')
            wano_path = os.path.join(self.simstack_path, 'wanos', self.routine_dir[i])
            if not os.path.exists(wano_path):
                os.makedirs(wano_path)
            xml_path = os.path.join(
                self.simstack_path,
                'wanos',
                self.routine_name[i],
                self.routine_name[i] +
                '.xml')    
            # Strat xml file generation
            file = open(xml_path, 'w')
            file.write('<WaNoTemplate>\n')
            file.write(" <WaNoRoot name='" + self.routine_name[i] + "'>\n")

            for j in range(self.subroutine_number[i]):
                for k in range(len(self.inputs_nodup[i][j])):
                    if self.inputs_nodup[i][j][k] != \
                            'duplicate' and self.inputs_no_locdep[i][j][k] != 'inner_dependent'\
                            and self.inputs_links[i][j][k][0].isnumeric() == False:
                        file.write(
                            "\t<WaNoFile name='" +
                            self.inputs_locname[i][j][k] +
                            " Data' logical_filename='" +
                            self.inputs_nodup[i][j][k] +
                            ".yaml' local='True'>" +
                            self.args_inputpath +
                            self.inputs_nodup[i][j][k] +
                            ".yaml</WaNoFile> \n")
                    elif self.inputs_nodup[i][j][k] != \
                        'duplicate' and self.inputs_no_locdep[i][j][k] != 'inner_dependent'\
                            and self.inputs_links[i][j][k][0].isnumeric():
                        path_dependent_output = os.path.join(self.routine_name[int(self.inputs_links[i][j][k][0])], self.outputs_locname[int(
                            self.inputs_links[i][j][k][0])][int(self.inputs_links[i][j][k][1])][int(self.inputs_links[i][j][k][2])] + '.yaml')
                        file.write(
                            "\t<WaNoFile name='" +
                             self.inputs_locname[i][j][k] +
                            " Data' logical_filename='" +
                             self.inputs_locname[i][j][k] +
                            ".yaml' local='False'>" +
                            path_dependent_output +
                            "</WaNoFile> \n")

            file.write(' </WaNoRoot>\n')
            file.write(
                " <WaNoExecCommand> bash ./run_python.sh</WaNoExecCommand>\n")
            file.write(" <WaNoInputFiles>\n")
            file.write("\t<WaNoInputFile logical_filename='run_python.sh'>"
                        "run_python.sh</WaNoInputFile>\n")
            file.write(
                "\t<WaNoInputFile logical_filename='" +
                self.routine_name[i] +
                "_wrapper.py'>" +
                self.routine_name[i] +
                "_wrapper.py</WaNoInputFile>\n")
            file.write(
                "\t<WaNoInputFile logical_filename='setenv.sh'>setenv.sh</WaNoInputFile>\n")
            file.write(" </WaNoInputFiles>\n")
            file.write(" <WaNoOutputFiles>\n")
            for j in range(self.subroutine_number[i]):
                for k in range(len( self.outputs_nodup[i][j])):
                    if  self.outputs_nodup[i][j][k] != 'duplicate':
                        file.write(
                            "\t<WaNoOutputFile logical_filename='" +
                             self.outputs_locname[i][j][k] +
                            ".yaml'>" +
                             self.outputs_locname[i][j][k] +
                            ".yaml</WaNoOutputFile>\n")
            file.write(" </WaNoOutputFiles>\n")
            file.write("</WaNoTemplate>")

            file.close()

            # Validate xml against schema
            schema_path = os.path.join(
                os.path.dirname(
                    os.path.abspath(__file__)),
                'xmlschema.xsd')
            my_schema = xmlschema.XMLSchema(schema_path)
            my_schema.validate(xml_path)

    def workflow_generator(self):
        
        simstack_wf_path = os.path.join(self.simstack_path, 'simstack.xml')
        file = open(simstack_wf_path, 'w')
        file.write('<root> \n')
        for i in range(self.routine_number):
            file.write(
                "\t <WaNo type='" +
                self.routine_name[i] +
                "' uuid='" +
                self.routine_name[i] +
                "' id='" +
                str(i) +
                "' name='" +
                self.routine_name[i] +
                "'/> \n")
        file.write('</root> \n')

        for i in range(self.routine_number):
            run_python_path = os.path.join(
                self.simstack_path, 'wanos', self.routine_name[i], 'run_python.sh')
            file = open(run_python_path, 'w')
            file.write("#!/bin/bash/ -ex \n")
            file.write("source setenv.sh \n")
            file.write("python " + self.routine_dir[i] + "_wrapper.py")

            file.close()

    def pack_wano(self):

        for i in range(self.routine_number):
            dest_path = os.path.join(
                self.simstack_path, 'wanos', self.routine_dir[i])
            source_path = os.path.join(
                os.path.dirname(
                    os.path.abspath(__file__)),
                'setenv.sh ')
            cmd = 'cp -r ' + source_path + dest_path+'/'
            os.popen(cmd)
            source_path = os.path.join(
                os.path.dirname(
                    os.path.abspath(__file__)),
                '..',
                'fig',
                'wano_fig.png ')
            cmd = 'cp ' + source_path + dest_path
            os.popen(cmd)
            source_path = os.path.join(
                os.path.dirname(
                    os.path.abspath(__file__)),
                'resources.yml ')
            cmd = 'cp ' + source_path + dest_path
            os.popen(cmd)

    # End of WaNo Folder generation


import os
import os.path
import sys
from pathlib import Path
import time
import json
import yaml
from ipywidgets import Button, HBox, VBox, Layout, ToggleButton
from IPython.display import HTML, display, clear_output, Markdown as md, JSON, Javascript, FileLink, Image, display_javascript, display_html, display_json
import ipywidgets as widgets
from ipywidgets import Layout
from ipyfilechooser import FileChooser
from json2html import *
from fabric2 import Connection
import subprocess
from wfgenes.core.BUILTIN import transfer_files
import site
from wfgenes import wfgenes_gui


wfgenes_path = os.path.join(site.getsitepackages()[0], 'wfgenes')
if wfgenes_gui.default_path.selected_path is not None:  
    default_path = wfgenes_gui.default_path.selected_path
else:
    default_path = wfgenes_path    


wfgenes_path = os.path.join(site.getsitepackages()[0], 'wfgenes')
fig_path = os.path.join(wfgenes_path, 'fig')
info_path = os.path.join(wfgenes_path, 'info')


text_input_layout = Layout(width='90%')

style = {'description_width': 'initial'}

# Choose python model's path
model_path = FileChooser('')
model_path.default_path = default_path
model_path.default_filename = ''
model_path.use_dir_icons = True
model_path.filter_pattern = ['*.py']
model_path.title = '<b>Path to the python model:</b>'


plus_lib = widgets.Button(
    tooltip='Transfer additional module(s) to remote machines',
    icon='plus',
    style=style,
    layout=Layout(
        width='100px'),
    description='Modules',
    title='add modules')
plus_lib_output = widgets.Output()

refresh_lib_button = widgets.Button(
    tooltip='Clear module list',
    icon='refresh',
    style=style,
    layout=Layout(
        width='40px'),
    description='')
refresh_lib_button_output = widgets.Output()


input_button = widgets.Button(
    tooltip='Locate inputs(s)',
    icon='plus',
    style=style,
    layout=Layout(
        width='100px'),
    description='Inputs')
input_button_output = widgets.Output()

refresh_input_button = widgets.Button(
    tooltip='Clear input list',
    icon='refresh',
    style=style,
    layout=Layout(
        width='40px'),
    description='')
refresh_input_button_output = widgets.Output()


info_button = widgets.ToggleButton(
    tooltip='Show/Hide Hint',
    icon='info',
    style=style,
    layout=Layout(
        width='60px'),
    description='Hint')
info_button_output = widgets.Output()


user = widgets.Textarea(
    value='th7356',
    placeholder='xy1234',
    description='Username',
    tooltip='Username on the remote cluster',
    layout=text_input_layout,
    disabled=False)

host = widgets.Textarea(
    value='horeka.scc.kit.edu',
    placeholder='hostname.domainname',
    description='Hostname',
    tooltip='Hostname of the remote cluster',
    layout=text_input_layout,
    disabled=False)

pre_launch_command = widgets.Textarea(
    value='',
    placeholder='module load xyz or Conda activate myenv',
    description='Command : Setup the environment before submitting job',
    tooltip='Remote configuration command',
    layout=text_input_layout,
    disabled=False)

launchdir = widgets.Textarea(
    value='/home/hk-project-test-sdlmat/th7356/test_wfgenes',
    placeholder='/absolute/path/to/launchdir',
    description='Launchdir',
    tooltip='Directory in that nodes will be executed',
    layout=text_input_layout
)


syspath_textarea = widgets.Textarea(
    placeholder='Type something',
    description='<b>Path file:</b>',
    description_tooltip='Edit to modify the graph',
    disabled=False,
    style=style,
    layout=Layout(width='100%', height='1000px')
)

pool_type = widgets.Dropdown(
    options=['local_threads', 'local_processes', 'slurm'],
    value='local_threads',
    description='<b>Pool type:<b>',
    disabled=False,
    style=style
)

pool_type_output = widgets.Output()

walltime = widgets.Text(
    value='00:30:00',
    description='<b>Wall time(queue)<b>',
    disabled=False,
    style=style
)

partition = widgets.Text(
    value='dev_cpuonly',
    placeholder='',
    description='<b>Partition(queue)<b>',
    disabled=False,
    style=style
)


memory = widgets.BoundedIntText(
    value=100,
    min=30,
    max=20000,
    step=1,
    description='<b> Total memory per job (GB): <b>',
    disabled=False,
    style=style
)


scale_nodes = widgets.BoundedIntText(
    value=1,
    min=1,
    step=1,
    description='<b>Number of nodes:<b>',
    disabled=False,
    style=style
)

workers_nodes = widgets.BoundedIntText(
    value=1,
    min=1,
    max=10000,
    step=1,
    description='<b>Workers(Processes) per node:<b>',
    disabled=False,
    style=style
)

cpu_per_node = widgets.BoundedIntText(
    value=1,
    min=1,
    max=1000,
    step=1,
    description='<b>CPUs per node:<b>',
    disabled=False,
    style=style
)


threads_number = widgets.BoundedIntText(
    value=1,
    min=1,
    step=1,
    description='<b>Maximum local threads:<b>',
    disabled=False,
    style=style
)
threads_number_output = widgets.Output()

processes_number = widgets.BoundedIntText(
    value=1,
    min=1,
    step=1,
    description='<b>Maximum number of parallel process:<b>',
    disabled=False,
    style=style
)
processes_number_output = widgets.Output()

cpu_per_process = widgets.BoundedIntText(
    value=1,
    min=1,
    step=1,
    description='<b>Number of CPU for each process :<b>',
    disabled=False,
    style=style
)
cpu_per_process_output = widgets.Output()


# Button RUN MODEL
run_model = widgets.Button(
    tooltip='Run model',
    icon='play',
    description='Run model',
    style=style)
output1 = widgets.Output()


# A python program to create user-defined exception

# class MyError is derived from super class Exception
class InputError(Exception):

    # Constructor or Initializer
    def __init__(self):
        self.message = 'Please Select one inputpath. To use local executor you need to have only one input path which will be used as a working directory.'

    # __str__ is to print() the value
    def __str__(self):
        return(repr(self.message))




def set_local_env(libpath, launch_comm):
    # export necessary PYTHONPATH
    if len(libpath) != 0:
        for path in libpath:
            if path.selected_path is not None:
                export_path = 'export PYTHONPATH=$PYTHONPATH:' + path.selected_path
                launch_comm.extend([export_path])


def set_cwd(input_path, launch_comm, pre_launch_command):
    # Set current working directory of local executor
    try:
        if len(input_path) != 1:
            raise(InputError())
        elif input_path[0].selected_path is not None:
            cd_to_cwd = 'cd ' + input_path[0].selected_path
            launch_comm.extend([cd_to_cwd])
        else:
            raise(MyError())
        if pre_launch_command:
            launch_comm.extend([pre_launch_command])
        else:
            pass
    except InputError as error:
        print(error.message)
        sys.exit()


libpath = []
def plus_lib_clicked(b3):
    libpath.append(FileChooser(''))
    libpath[len(libpath) - 1].default_path = default_path
    libpath[len(libpath) - 1].use_dir_icons = True
    with plus_lib_output:
        if len(libpath) == 1:
            display(refresh_lib_button)
        display(libpath[len(libpath) - 1])


input_path = []
def input_button_clicked(b3):
    input_path.append(FileChooser(''))
    input_path[len(input_path) - 1].default_path = default_path
    input_path[len(input_path) - 1].use_dir_icons = True
    with input_button_output:
        if len(input_path) == 1:
            display(refresh_input_button)
        display(input_path[len(input_path) - 1])


def refresh_input_button_clicked(b):
    with input_button_output:
        clear_output()
        input_path.clear()


def refresh_lib_button_clicked(b):
    with plus_lib_output:
        clear_output()
        libpath.clear()


def info_button_clicked(obj):
    with info_button_output:
        if obj['new']:
            slurm_single = os.path.join(fig_path, 'slurm_single.png')
            slurm_multi = os.path.join(fig_path, 'slurm_multi.png')
            blank = os.path.join(fig_path, 'blank.png')
            img1 = open(slurm_single, 'rb').read()
            wi1 = widgets.Image(
                value=img1,
                format='jpg',
                width=300,
                height=400)
            img0 = open(blank, 'rb').read()
            wi0 = widgets.Image(
                value=img0,
                format='jpg',
                width=300,
                height=405)
            img2 = open(slurm_multi, 'rb').read()
            wi2 = widgets.Image(
                value=img2,
                format='jpg',
                width=300,
                height=405)
            a = [wi1, wi0, wi2]
            wid = widgets.HBox(a)
            display(wid)
        else:
            with info_button_output:
                clear_output()


def run_model_clicked(b):
    with output1:
        # Clear cell output
        for i in range(10):
            clear_output(wait=True)
        launch_comm = []    # launch command
        argument_threads = ["--pool", "--threads_number"]
        value_threads = [pool_type.value, threads_number.value]

        argument_processes = [
            "--pool",
            "--processes_number",
            "--cpu_per_process"]
        value_processes = [
            pool_type.value,
            processes_number.value,
            cpu_per_process.value]

        argument_slurm = [
            "--pool",
            "--partition",
            "--walltime",
            "--memory",
            "--scale",
            "--worker_per_node",
            "--cpu_per_node"]
        value_slurm = [
            pool_type.value,
            partition.value,
            walltime.value,
            memory.value,
            scale_nodes.value,
            workers_nodes.value,
            cpu_per_node.value]

        if pool_type.value == 'local_threads':
            argument_list = argument_threads
            value_list = value_threads
            launch_model = "python " + str(model_path.selected)
        elif pool_type.value == 'local_processes':
            argument_list = argument_processes
            value_list = value_processes
            launch_model = "python " + str(model_path.selected)
        elif pool_type.value == 'slurm':
            argument_list = argument_slurm
            value_list = value_slurm
            launch_model = "python " + str(model_path.selected_filename)

        for argument, value in zip(argument_list, value_list):
            launch_model += " " + argument + " " + str(value)

        if pool_type.value == 'local_processes' or pool_type.value == 'local_threads':
            set_local_env(libpath, launch_comm)
            set_cwd(input_path, launch_comm, pre_launch_command.value)
            launch_comm.extend([launch_model])
            cmd = '&&'.join(launch_comm)
            print(cmd)
            proc = subprocess.call(cmd, shell=True)

        elif pool_type.value == 'slurm':
            with Connection(host=host.value, user=user.value) as conn:
                conn.run('mkdir -p ' + launchdir.value)
                transfer_files(user.value, host.value, input_path, launchdir.value)
                conn.put(model_path.selected, launchdir.value)
                launch_comm.extend(
                    [transfer_files(user.value, host.value, libpath, launchdir.value, 'lib')])
                with conn.cd(launchdir.value):
                    if model_path.selected_filename is not None:
                        if pre_launch_command.value:
                            launch_comm.extend(
                                [pre_launch_command.value, launch_model])
                        else:
                            launch_comm.extend([launch_model])
                        cmd = '&&'.join(launch_comm)
                        print(cmd)
                        conn.run(cmd)
                    else:
                        print('Select your python model')


def pool_type_changed(bvar):
    """ Observe Pool type"""
    with pool_type_output:
        clear_output()
        if pool_type.value == 'local_threads':
            display(threads_number)
        elif pool_type.value == 'local_processes':
            display(HBox([processes_number, cpu_per_process]))
        elif pool_type.value == 'slurm':
            display(user, host)
            display(HBox([partition, walltime, memory]))
            display(HBox([cpu_per_node, workers_nodes, scale_nodes]))
            display(info_button, info_button_output)
            display(launchdir)


# edit_environment.on_click(edit_environment_clicked)
pool_type.observe(pool_type_changed)
plus_lib.on_click(plus_lib_clicked)
input_button.on_click(input_button_clicked)
run_model.on_click(run_model_clicked)
refresh_input_button.on_click(refresh_input_button_clicked)
refresh_lib_button.on_click(refresh_lib_button_clicked)
info_button.observe(info_button_clicked, 'value')


def display_executor():
    """ Dask/Parsl Executor Top View"""
    display(model_path)
    display(input_button, input_button_output)
    display(pre_launch_command)
    display(plus_lib, plus_lib_output)
    display(pool_type, pool_type_output)
    pool_type_changed('NULL')
    display(HBox([run_model]), output1)



import os
import os.path
from copy import deepcopy
from collections import OrderedDict
import shutil
import argparse
import json
import yaml
from pathlib import Path
import random
ROOT_DIR = os.path.dirname(os.path.abspath(__file__)) 
wconfig = {}
dag_file = 'rdag_10.yaml'
graph_order = 10
randomlist_size = int(graph_order / 10)

randomlist = []
i = 0
while i < randomlist_size:
    n = random.randint(1, graph_order)
    i +=1
    if n not in randomlist:
        randomlist.append(n)
    else:
        i -=1    
print(randomlist)


with open(dag_file) as input_stream:
    random_dag = yaml.load(input_stream, Loader=yaml.Loader)
graph_edges = random_dag['graph']['edges']
graph_nodes = random_dag['graph']['nodes']
num_nodes = len(random_dag['graph']['nodes'])
num_nodes +=1 # Last node- target node only for nodes without target  
graph_nodes.append({'id': str(num_nodes)})
wconfig = {}
nodes_wconfig = []
no_output_nodes = []
for node_graph in graph_nodes:
    node = {}
    node['tasks']= []
    node['name'] = 'node_id'+ node_graph['id']
    node['id'] = int(node_graph['id'])
    task = {}
    task['func'] = []
    task['inputs'] = []
    task['outputs'] = []
    task['func'].append('my_function')    
    task['func'].append('weighted_sleep')
    for edge in graph_edges:
        if edge['source'] == node_graph['id']:              
            task['outputs'].append('link_'+ edge['source']+'_to_'+ edge['target']) 
        elif edge['target'] == node_graph['id']:
            task['inputs'].append('link_' + edge['source']+'_to_'+ edge['target'])
    if (len(graph_edges) == 0 or len(task['outputs']) == 0) and node_graph['id'] != str(num_nodes):
        task['outputs'].append('link_'+ node_graph['id']+'_to_'+ str(num_nodes))
        no_output_nodes.append(node_graph['id'])
    if node_graph['id'] == str(num_nodes):
        for source in no_output_nodes:
            task['inputs'].append('link_' + str(source)+ '_to_' +  node_graph['id'])
    task['kwargs'] = {}
    if int(node_graph['id']) in randomlist:
        task['kwargs'] = { 'len_output' : len(task['outputs']), 'sleep_time': 100, 'id': int(node_graph['id'])}
    else:
        task['kwargs'] = { 'len_output' : len(task['outputs']), 'sleep_time': 10, 'id': int(node_graph['id'])}
    node['tasks'].append(task)
    nodes_wconfig.append(node)


workflow_name = str('node_') + str(len(graph_nodes))+ '_' + str(len(graph_edges))
wconfig = {'workflow_name': workflow_name, 'nodes': nodes_wconfig }
noalias_dumper = yaml.dumper.SafeDumper
noalias_dumper.ignore_aliases = lambda self, data: True
with open('wconfig_unbalanced.yaml', 'w') as output_stream:
    yaml.dump(wconfig, output_stream)      




"""Launch workflow nodes on remote resources"""

import os, sys
import datetime
from time import sleep
from fabric2 import Connection
from fireworks.fw_config import FW_BLOCK_FORMAT
from fireworks import LaunchPad, FWorker
from fireworks.utilities.fw_serializers import recursive_serialize
from fireworks.utilities.fw_serializers import recursive_deserialize
from wfgenes.engine.wfengine import WFEngine
from wfgenes.core.BUILTIN import transfer_files


class WFEngineRemote(WFEngine):
    """
    A subclass of wfEngine to manage remote workers

    Args:
        host: hostname of the remote resource
        user: username on the remote resource
        conf: configuration command to set up the remote environment

    Passwordless connection via SSH to the remote system must be enabled.
    Otherwise the following error message will occur:
        PasswordRequiredException: private key file is encrypted
    """

    def __init__(self, launchpad, qadapter, wf_query, host=None, user=None,
                 conf='', libpath=[] ,**kwargs):
        super().__init__(launchpad, qadapter, wf_query, **kwargs)
        assert host and user and qadapter
        self.host = host
        self.user = user
        self.conf = conf  # for example '. python-3.6.8-custom/bin/activate'
        self.libpath = libpath
        self.setup_remote_launchpad()
        self.setup_remote_qadapter()
        self.setup_remote_fworker()
        self.setup_remote_configuration()

    def setup_remote_launchpad(self):
        """Create launchpad file for remote worker"""
        lpad = self.launchpad
        self.remote_launchpad_file = 'launchpad_' + self.name + '.yaml'
        remote_launchpad = LaunchPad.from_dict(lpad.to_db_dict())
        self.remote_launchpad_path = os.path.join(self.launchdir, self.remote_launchpad_file)
        if hasattr(lpad, 'ssl_ca_certs') and lpad.ssl_ca_certs:
            ssl_ca_certs_file = os.path.basename(lpad.ssl_ca_certs)
            ssl_ca_certs_path = os.path.join(self.launchdir, ssl_ca_certs_file)
            remote_launchpad.ssl_ca_certs = ssl_ca_certs_path
        if hasattr(lpad, 'ssl_certfile') and lpad.ssl_certfile:
            ssl_certfile_file = os.path.basename(lpad.ssl_certfile)
            ssl_certfile_path = os.path.join(self.launchdir, ssl_certfile_file)
            remote_launchpad.ssl_certfile = ssl_certfile_path
        remote_launchpad.to_file(self.remote_launchpad_file)
        lp_mongo_new = lpad.mongoclient_kwargs
        if (hasattr(lpad, 'mongoclient_kwargs') and
                'tlsCAFile' in lp_mongo_new and lp_mongo_new['tlsCAFile']):
            tlscafile = os.path.basename(lp_mongo_new['tlsCAFile'])
            tlscapath = os.path.join(self.launchdir, tlscafile)
            remote_launchpad.mongoclient_kwargs['tlsCAFile'] = tlscapath
        if (hasattr(lpad, 'mongoclient_kwargs') and
                'tlsCertificateKeyFile' in lp_mongo_new and
                lp_mongo_new['tlsCertificateKeyFile']):
            tlscrlfile = os.path.basename(lp_mongo_new['tlsCertificateKeyFile'])
            tlscrlpath = os.path.join(self.launchdir, tlscrlfile)
            remote_launchpad.mongoclient_kwargs['tlsCertificateKeyFile'] = tlscrlpath
        remote_launchpad.to_file(self.remote_launchpad_file)

    def setup_remote_qadapter(self):
        """Create qadapter file for remote worker"""
        self.qadapter_file = 'qadapter_' + self.name + '.yaml'
        self.remote_qadapter_path = os.path.join(self.launchdir, self.qadapter_file)
        if 'pre_rocket' in self.qadapter and self.qadapter['pre_rocket'] is not None:
            commands = self.qadapter['pre_rocket'].split(';')  # List commands of pre_rocket
            if commands[-1] != self.conf:  # Check if the pre_rocket is already modified
                self.qadapter['pre_rocket'] = self.qadapter['pre_rocket'] + ';' + self.conf
        else:
            self.qadapter['pre_rocket'] = self.conf
        self.qadapter['rocket_launch'] = 'rlaunch -l ' + self.remote_launchpad_path + ' singleshot'
        self.qadapter.to_file(self.qadapter_file)

    def setup_remote_fworker(self):
        """Create configuration for remote worker"""
        self.fworker_file = 'fworker_' + self.name + '.yaml'
        self.remote_fworker_path = os.path.join(self.launchdir, self.fworker_file)
        FWorker(name=self.name, category='remote').to_file(self.fworker_file)

    def setup_remote_configuration(self):
        """Create remote launch directory and copy all configuration files"""
        with Connection(host=self.host, user=self.user) as conn:
            conn.run('mkdir -p ' + self.launchdir)
            conn.put(local=self.remote_launchpad_file, remote=self.launchdir)
            conn.put(local=self.qadapter_file, remote=self.launchdir)
            conn.put(local=self.fworker_file, remote=self.launchdir)
            if hasattr(self.launchpad, 'ssl_ca_certs') and self.launchpad.ssl_ca_certs:
                conn.put(local=self.launchpad.ssl_ca_certs, remote=self.launchdir)
            if hasattr(self.launchpad, 'ssl_certfile') and self.launchpad.ssl_certfile:
                conn.put(local=self.launchpad.ssl_certfile, remote=self.launchdir)

    def slaunch(self, fw_id):
        """Launch a batch node on a remote resource"""
        # do some tests in case this is not called by launcher()
        #assert fw_id in self.get_fw_ids(), 'invalid fw_id'
        firework = self.launchpad.get_fw_by_id(fw_id)
        last_state = firework.archived_launches[-1].state
        #assert firework.state == 'READY', 'invalid state'
        fw_spec = firework.spec
        #assert '_category' in fw_spec, 'no category specified'
        #assert '_fworker' in fw_spec, 'no worker specified'
        #assert fw_spec['_category'] == 'remote', 'invalid category'
        #assert fw_spec['_fworker'] == self.name, 'invalid worker name'
        
        #if firework.state != 'FIZZLED':
        prefix = 'launcher_'
        uniq_name = prefix+datetime.datetime.utcnow().strftime(FW_BLOCK_FORMAT)
        new_launch_dir = os.path.join(self.launchdir, uniq_name)
        #else    
        launch_comm = []
        if self.conf:
            launch_comm.extend([self.conf, '&&'])
        launch_comm.extend([transfer_files(self.user, self.host, self.libpath, self.launchdir, 'lib'), '&&'])
        launch_comm.extend(['mkdir -p', new_launch_dir])
        launch_comm.extend(['&&', 'qlaunch',
                            '--launchpad_file', self.remote_launchpad_path,
                            '--queueadapter_file', self.remote_qadapter_path,
                            '--fworker_file', self.remote_fworker_path,
                            #'--launch_dir', new_launch_dir])
                            '--launch_dir', self.launchdir])
        launch_comm.extend(['--reserve', 'singleshot', '--fw_id', str(fw_id)])
        print(' '.join(launch_comm))
        sys.stdout.flush()
        with Connection(self.host, self.user) as conn:
            conn.run(' '.join(launch_comm))

    def launcher(self, stop_event):
        """Awake every sleep_time seconds and launch all READY nodes"""
        wf_query = {'nodes': {'$in': self.wf_ids}}
        fw_query = {'state': 'READY', 'spec._category': 'remote',
                    'spec._fworker': self.name}
        while not stop_event.is_set():
            fw_ids = self.launchpad.get_fw_ids_in_wfs(wf_query, fw_query)
            if not fw_ids:
                print('There is no READY task to launch')
                sys.stdout.flush()
            for fw_id in fw_ids:
                self.slaunch(fw_id)
            sleep(self.sleep_time)
        print('launcher thread stopped')

    def exec_cancel(self, res_id):
        """Execute the slurm cancel command remotely"""
        with Connection(host=self.host, user=self.user) as conn:
            conn.run(f'scancel {res_id}')

    def check_jobcancel(self, res_id):
        """Execute the slurm sacct command remotely"""
        with Connection(host=self.host, user=self.user) as conn:
            output = conn.run(f'sacct -j {res_id}')
            substring = 'CANCELLED'
            if substring not in output.stdout:
                raise RuntimeError('Error job is not cancelled')

    @recursive_serialize
    def to_dict(self):
        """
        Serialize the engine object to a dictionary

        Returns:
            (dict): a dictionary with all parameters needed to call __init__
        """
        m_dict = super().to_dict()
        m_dict.update({'host': self.host, 'user': self.user, 'conf': self.conf})
        return m_dict

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        """
        Construct an engine object from a dictionary

        Args:
            m_dict (dict): a dict with all parameters needed to call __init__

        Returns:
            FWEngineRemote object
        """
        launchpad = LaunchPad.from_dict(m_dict.get('launchpad'))
        qadapter = m_dict.get('qadapter')
        wf_query = m_dict.get('wf_query')
        name = m_dict.get('name')
        launchdir = m_dict.get('launchdir')
        sleep_time = m_dict.get('sleep_time')
        host = m_dict.get('host')
        user = m_dict.get('user')
        conf = m_dict.get('conf')
        return cls(launchpad=launchpad, qadapter=qadapter, wf_query=wf_query,
                   name=name, launchdir=launchdir, sleep_time=sleep_time,
                   host=host, user=user, conf=conf)



""" Builtin modules for wfGenes """
import numpy as np

def merge_dic(*files):
    
    file_update = {}
    for file in files:
        if isinstance(file, np.number):
            file = dict(file = float(file))
        else:
            file = dict(file = file)
        file_update = {**file, **file_update}
    return file_update
      


#from parsl import python_app
import time
import yaml


#@python_app
def data_generator(list_length):
    # Generate and initialize list 
    data = [0 for i in range(list_length)]
    for i in range(list_length):
        data[i]= i +1
        with open('data.yaml', 'w') as outfile:
            yaml.dump(data, outfile, default_flow_style=False)
    return data    


#@python_app
def plus_minus(x, y):
    """ x - a list, y - a number """
    return_list_x = []
    return_list_y = []
    for i in x:
        return_list_x.append(i + y )
        return_list_y.append(i - y )
        time.sleep(1)
    return return_list_x, return_list_y

#@python_app
def sum_square(x, y, z):
    """ x, y - lists of same length, z - a number """
    return_list = []
    for i, j in zip(x, y):
        return_list.append(z*(i**2+j**2))
        time.sleep(1)
    return return_list

#@python_app
def average_list(x):
    """ x, y - lists of same length, z - a number """
    total_sum = 0
    for i in x:
        total_sum = total_sum + i
    return [total_sum / len(x)]    

if __name__ == '__main__':
    data = [1, 2, 3, 4]
    datax, datay = plus_minus(data, 1)
    results = sum_square(datax, datay, 2)
    print(results)



""" classes and functions for working with atoms using ASE """

__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2018, Karlsruhe Institute of Technology'

import numpy as np
from igraph import Graph
from ase.neighborlist import NeighborList
from ase.neighborlist import neighbor_list

def collinear(atom1, atom2, atom3):
    """ test collinearity of three atoms """
    vec1 = atom1.position - atom2.position
    vec2 = atom1.position - atom3.position
    tol = np.sqrt(np.finfo(np.float).eps)
    return np.linalg.norm(np.cross(vec1, vec2)) < tol


def make_graph_dict(atoms, cutoff=1.4, method='n1'):
    """ linearly (n1) and quadratically (n2) scaling algorithm """
    if method == 'n1':
        cutoff = 2.1*cutoff
        els = {tuple(sorted(t)) for t in zip(*neighbor_list('ij', atoms, cutoff))}
    if method == 'n2':
        nl = NeighborList(cutoffs=len(atoms)*[cutoff], bothways=False,
                          self_interaction=False)
        nl.update(atoms)
        els = [(a.index, n) for a in atoms for n in nl.get_neighbors(a.index)[0]]
    return {'n': len(atoms), 'edges': list(els), 'vertex_attrs': {'id': list(range(len(atoms)))}}


def make_graph(atoms, cutoff=1.4, method='n1'):
    """ alternative to AtomsGraph class for faster pickle/unpickle """
    return Graph(**make_graph_dict(atoms, cutoff, method))


class AtomsGraph(Graph):
    """ construct a graph for a structure in an Atoms object """
    def __init__(self, atoms, cutoff=1.4, method='n1'):
        self.args = (atoms, cutoff, method)
        graph_dict = make_graph_dict(atoms, cutoff, method)
        super(AtomsGraph, self).__init__(**graph_dict)

    def __reduce__(self):
        info = super(AtomsGraph, self).__reduce__()
        return (info[0], self.args)



# -*- coding: utf-8 -*-

import imutils
import cv2
import numpy as np
import time
import dask
#from parsl import python_app
# =============================================================================
# USER-SET PARAMETERS
# =============================================================================

# Number of frames to pass before changing the frame to compare the current
# frame against
FRAMES_TO_PERSIST = 10

# Minimum boxed area for a detected motion to count as actual motion
# Use to filter out noise or small objects
MIN_SIZE_FOR_MOVEMENT = 2000

# Minimum length of time where no motion is detected it should take
#(in program cycles) for the program to declare that there is no movement
MOVEMENT_DETECTED_PERSISTENCE = 100

# =============================================================================
# CORE PROGRAM
# =============================================================================


#@python_app
def count_cameras():
    """ count number of connected cameras and extract their index """
    max_tested = 10
    cameras_count = 0
    cameras_index = []
    for i in range(max_tested):
        try:
            temp_camera = cv2.VideoCapture(i)
            if temp_camera.isOpened():
                temp_camera.release()
                cameras_count +=1
                cameras_index.append(i)
                continue
        except:
            print('pass')    
    return cameras_count, cameras_index

#@python_app
def motion_control(index_list):
    """
    Detect motion in indexed camera
    References/Adapted From: 
    https://www.pyimagesearch.com/2015/05/25/basic-motion-detection-and-tracking-with-python-and-opencv/
    github.com/methylDragon
    """
    # Init frame variables
    first_frame = None
    next_frame = None
    
    # extract the index from index_list
    index = index_list [0] 

    # Init display font and timeout counters
    font = cv2.FONT_HERSHEY_SIMPLEX
    delay_counter = 0
    movement_persistent_counter = 0
    
    cap = cv2.VideoCapture(index) # Flush the stream
    cap.release()
    cap = cv2.VideoCapture(index) # Then start the webcam

    # LOOP!
    t_end = time.time() + 20
    #while time.time() < t_end:
    while True:  
        # Set transient motion detected as false
        transient_movement_flag = False
        
        # Read frame
        ret, frame = cap.read()
        text = "Unoccupied"

        # If there's an error in capturing
        if not ret:
            print("CAPTURE ERROR")
            continue

        # Resize and save a greyscale version of the image
        frame = imutils.resize(frame, width = 750)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # Blur it to remove camera noise (reducing false positives)
        gray = cv2.GaussianBlur(gray, (21, 21), 0)

        # If the first frame is nothing, initialise it
        if first_frame is None: first_frame = gray    

        delay_counter += 1

        # Otherwise, set the first frame to compare as the previous frame
        # But only if the counter reaches the appriopriate value
        # The delay is to allow relatively slow motions to be counted as large
        # motions if they're spread out far enough
        if delay_counter > FRAMES_TO_PERSIST:
            delay_counter = 0
            first_frame = next_frame

            
        # Set the next frame to compare (the current frame)
        next_frame = gray

        # Compare the two frames, find the difference
        frame_delta = cv2.absdiff(first_frame, next_frame)
        thresh = cv2.threshold(frame_delta, 25, 255, cv2.THRESH_BINARY)[1]

        # Fill in holes via dilate(), and find contours of the thesholds
        thresh = cv2.dilate(thresh, None, iterations = 2)
        cnts, _ = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        # loop over the contours
        for c in cnts:

            # Save the coordinates of all found contours
            (x, y, w, h) = cv2.boundingRect(c)
            
            # If the contour is too small, ignore it, otherwise, there's transient
            # movement
            if cv2.contourArea(c) > MIN_SIZE_FOR_MOVEMENT:
                transient_movement_flag = True
                
                # Draw a rectangle around big enough movements
                cv2.rectangle(frame, (x, y), (x + w, y + h), (0, 255, 0), 2)

        # The moment something moves momentarily, reset the persistent
        # movement timer.
        if transient_movement_flag == True:
            movement_persistent_flag = True
            movement_persistent_counter = MOVEMENT_DETECTED_PERSISTENCE

        # As long as there was a recent transient movement, say a movement
        # was detected    
        if movement_persistent_counter > 0:
            text = "Movement Detected " + str(movement_persistent_counter)
            movement_persistent_counter -= 1
        else:
            text = "No Movement Detected"

        # Print the text on the screen, and display the raw and processed video 
        # feeds
        cv2.putText(frame, str(text), (10,35), font, 0.75, (255,255,255), 2, cv2.LINE_AA)
        
        # For if you want to show the individual video frames
    #    cv2.imshow("frame", frame)
    #    cv2.imshow("delta", frame_delta)
        
        # Convert the frame_delta to color for splicing
        frame_delta = cv2.cvtColor(frame_delta, cv2.COLOR_GRAY2BGR)

        # Splice the two video frames together to make one long horizontal one
        window_name = str("frame_" + str(index))
        cv2.imshow(window_name, np.hstack((frame_delta, frame)))


        # Interrupt trigger by pressing q to quit the open CV program
        ch = cv2.waitKey(1)
        if ch & 0xFF == ord('q'):
            break

    # Cleanup when closed
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    cap.release()

if __name__ == "__main__":
    
    # Create capture object
    camera_count, cameras_index = count_cameras()
    print(camera_count, cameras_index)
    run_parallel = []
    for index in cameras_index:
        cam = dask.delayed(motion_control)(index)
        run_parallel.append(cam)
        
    dask.compute(*run_parallel)    
        
        
    
    



""" wfGenes: Automatic workflow generator."""


__author__ = 'Mehdi Roozmeh'
__email__ = 'mehdi.roozmeh@kit.edu'
__copyright__ = 'Copyright 2020, Karlsruhe Institute of Technology'


import os
import os.path
import sys
from copy import deepcopy
from collections import OrderedDict
import argparse
import json
import yaml
import fireworks_schema
from wfgenes.core.initial_wfgenes import BasewfGenes




class FireworkwfGenes():
    """ Automatic Fireworks generation from wconfig file """
    def __init__(self, wfgenes_init):
        self.firework_generation(wfgenes_init)
        self.workflow_generation(wfgenes_init)

    def firework_generation(self, wfgenes_init):
        self.fws = []
        for i in range(wfgenes_init.routine_number):
            fw = {}
            fw['fw_id'] = i
            fw['name'] = wfgenes_init.routine_name[i]
            fw['spec'] = {}
            fw['spec']['_tasks'] = []
            fw['spec']['_category'] = {}
            fw['spec']['_category'] = 'batch'
            
            #fw['spec']['_category'].append('batch')
            output_subroutine_iter = iter(wfgenes_init.outputs_locname[i])
            input_subroutine_iter = iter(wfgenes_init.inputs_locname[i])
            kwargs_subroutine_iter = iter(wfgenes_init.kwargs[i])
            resource_subroutine_iter = iter(wfgenes_init.resource[i])
            subroutine_iter = iter(wfgenes_init.func[i])
            # iterate over the tasklist
            for j in range(wfgenes_init.subroutine_number[i]):
                task = {}
                subroutine = next(subroutine_iter)
                kwargs = next(kwargs_subroutine_iter)
                resource = next(resource_subroutine_iter)
                if subroutine[0] != 'BUILTIN' and len(subroutine) <= 2:
                    task['_fw_name'] = 'PyTask'
                    task['func'] = str(subroutine[0]) + \
                        '.' + str(subroutine[1])
                elif subroutine[0] == 'BUILTIN' and subroutine[1] == 'MERGE':
                    task['_fw_name'] = 'JoinDictTask'
                elif subroutine[0] == 'BUILTIN' and subroutine[1] == 'callscript':
                    task['_fw_name'] = 'ScriptTask'
                    task['script'] = str (kwargs['command'])
                    if 'arguments' in kwargs:
                        task['script'] = task['script'] + ' '+ str (kwargs['arguments'])

                if len(subroutine) > 2 :
                    task['_fw_name'] = 'ForeachTask'
                    task['split'] = []
                    task['task'] = {}
                    task['split'] = subroutine[3]
                    if subroutine[4] == 'full':
                        pass
                    else:
                        task['number of chunks'] = []    
                        task['number of chunks'] = int(subroutine[4])
                    task['task']['_fw_name'] = []
                    task['task']['_fw_name'] = 'PyTask'
                    task['task']['func'] = str(subroutine[0]) + \
                        '.' + str(subroutine[1])

                inputs = next(input_subroutine_iter)
                if inputs and inputs[0] != 'NULL':   
                    if len(subroutine) == 2: # Normal task
                        task['inputs'] = []
                        task['inputs'] = inputs
                    elif len(subroutine) > 2 : # Data-flow Task. E.g foreach
                        task['task']['inputs'] = []
                        task['task']['inputs'] = inputs
                
                
                if bool(kwargs) and subroutine[0] != 'BUILTIN' and subroutine[1] != 'callscript' and kwargs != 'null' :   # Check if the dictionary is not empty
                    task['kwargs'] = {}
                    task['kwargs'] = kwargs
                if bool(resource) and subroutine[0] != 'BUILTIN' and subroutine[1] != 'callscript' and resource != 'null' :
                    if 'kwargs' not in task:
                        task['kwargs'] = {}
                        task['kwargs'] = resource
                    else:    
                        task['kwargs'].update(resource)
                    

                output = next(output_subroutine_iter)
                if len(output) > 0 and output[0] != 'NULL':
                    if subroutine[0] != 'BUILTIN' and len(subroutine) <= 2 :
                        task['outputs'] = []
                        task['outputs'] = output
                    elif subroutine[1] == 'MERGE' and len(subroutine) <= 2:
                        task['output'] = []
                        task['output'] = output[0]
                    elif len(subroutine) > 2:
                        task['task']['outputs'] = output          
                fw['spec']['_tasks'].append(task)
                # end of tasklist iterator

            for j in range(wfgenes_init.subroutine_number[i]):
                for k in range(len(wfgenes_init.inputs[i][j])):
                    if (wfgenes_init.inputs_no_locdep[i][j][k][0] != 'inner_dependent' and
                        wfgenes_init.inputs_nodup[i][j][k] != 'duplicate' and
                            wfgenes_init.inputs_links[i][j][k][0].isnumeric() == False and 
                            wfgenes_init.inputs_nodup[i][j][k] != 'NULL'):
                        input_string = os.path.join(
                            wfgenes_init.args_inputpath, str(
                                wfgenes_init.inputs[i][j][k]) + '.yaml')
                        with open(input_string, 'r') as input_stream:
                            fw['spec'][wfgenes_init.inputs_locname[i][j]
                                    [k]] = yaml.safe_load(input_stream) 
            self.fws.append(fw)

    def workflow_generation(self, wfgenes_init):
        firework_path = os.path.join(wfgenes_init.workflow_path, 'FireWorks')
        if not os.path.exists(firework_path):
            os.makedirs(firework_path)

        links = {}
        links_dot = {}

        for M in range(wfgenes_init.routine_number):
            links[str(M)] = []
            links_dot[str(M)] = []
            for i in range(wfgenes_init.routine_number):
                for j in range(wfgenes_init.subroutine_number[i]):
                    for k in range(len(wfgenes_init.inputs_links[i][j])):
                        if isinstance(
                                wfgenes_init.inputs_links[i][j][k],
                                list) and wfgenes_init.inputs_links[i][j][k][0] == str(M):
                            links[str(M)].append(i)
            links_dot[str(M)] = links[str(M)]
            links[str(M)] = list(OrderedDict.fromkeys(links[str(M)]))
        metadata = {}
        metadata = wfgenes_init.metadata

        workflow = {'fws': self.fws, 'links': links, 'metadata': metadata,
                    'name': wfgenes_init.interface_dict['workflow_name'] + '_wfGenes'}

#        fireworks_schema.validate(workflow, 'Workflow')
        pad_path = os.path.join(firework_path,
                                wfgenes_init.interface_dict['workflow_name'] + '.yaml')
        with open(pad_path, 'w') as output_stream:
            yaml.dump(workflow, output_stream)        



__author__ = 'Ivan Kondov'
__email__ = 'ivan.kondov@kit.edu'
__copyright__ = 'Copyright 2016, Karlsruhe Institute of Technology'


import re
import numpy as np
try: 
    from StringIO import StringIO 
except ImportError: 
    from io import StringIO 
from ase.io import read
from ase.units import Hartree
from ase.calculators.nwchem import NWChem

class KITnwchem(NWChem):
    def __init__(self, **parameters):

        self.set_parameters(parameters)
        assert self.verify_parameters() # need some handling

        NWChem.__init__(self,
            label=self.title,
            geometry=self.geometry,
            basis=self.basis_set,
            charge=self.charge,
            mult=self.multiplicity,
            xc=self.func,
            convergence=self.convergence,
            task=self.task
            )
        self.reset()

    def set_parameters(self, parameters):

        self.geometry = 'nocenter noautosym noautoz units angstroms'

        if 'command' in parameters.keys(): # e.g. "mpirun -np 4 nwchem PREFIX.nw > PREFIX.out"
            self.command = parameters['command']

        if 'title' in parameters.keys():
            self.title = parameters['title'].replace (" ", "_")
        else:
            self.title = 'notitle'
        if 'basis set' in parameters.keys():
            self.basis_set = parameters['basis set']
        if 'charge' in parameters.keys():
            self.charge = parameters['charge']
        else:
            self.charge = 0
        if 'multiplicity' in parameters.keys():
            self.multiplicity = parameters['multiplicity']
        if 'dft' in parameters.keys():
            self.dft = parameters['dft']
        if 'density functional' in parameters.keys():
            self.func = parameters['density functional']
        else:
            self.func = 'lda'
            print("Density functional set to default: "), self.func
        if 'ri' in parameters.keys():
            self.ri = parameters['ri']
        if 'internals' in parameters.keys():
            self.internals = parameters['internals']
        else:
            self.internals = False
        if 'scfiterlimit' in parameters.keys():
            self.scfiterlimit = parameters['scfiterlimit']
        else:
            self.scfiterlimit = 60
        if 'force convergence' in parameters.keys():
            self.force_conv = parameters['force convergence']
        else:
            self.force_conv = None
        if 'energy convergence' in parameters.keys():
            self.energy_conv = parameters['energy convergence']
        else:
            self.energy_conv = None
        self.convergence= {'energy': self.energy_conv, 'density': None, 'gradient': self.force_conv, 'lshift': None, 'damp': None}
        if 'task' in parameters.keys():
            self.task = parameters['task']
        else:
            self.task = 'energy' # note that 'gradient' is the default of the base class
        if 'geometry iterations' in parameters.keys():
            self.maxiter = parameters['geometry iterations']
        else:
            self.maxiter = None


    def verify_parameters(self):
        available_functionals = [ # only "parameter-free" funcionals from all supported fnctionals
            'acm', 'b3lyp', 'beckehandh',  'pbe0', 'becke97',   'becke97-1', 'becke97-2', 
            'becke97-3', 'becke97-d', 'becke98', 'hcth', 'hcth120', 'hcth147', 'hcth407', 
            'becke97gga1',  'hcth407p', 'mpw91', 'mpw1k', 'xft97', 'cft97', 'ft97', 'op', 
            'bop', 'pbeop', 'xpkzb99', 'cpkzb99', 'xtpss03', 'ctpss03', 'xctpssh', 'b1b95', 
            'bb1k', 'mpw1b95', 'mpwb1k', 'pw6b95', 'pwb6k', 'm05', 'm05-2x', 'vs98', 'm06',
            'm06-hf', 'm06-L', 'm06-2x'
            ]
        available_basis_sets = [ # no ECP, no check for availability for given elements
            '3-21++G', '3-21++G*', '3-21G', '3-21G*', '3-21GSP', '4-22GSP', '4-31G', '6-31++G', '6-31++G*', '6-31++G**', '6-31+G*', '6-311++G(2d,2p)', '6-311++G(3df,3pd)', '6-311++G**', '6-311+G*', '6-311G', '6-311G(2df,2pd)', '6-311G*', '6-311G**', '6-31G', '6-31G-Blaudeau', '6-31G(3df,3pd)', '6-31G*', '6-31G*-Blaudeau', '6-31G**', 'Ahlrichs pVDZ', 'Ahlrichs TZV', 'Ahlrichs VDZ', 'Ahlrichs VTZ', 'ANO-RCC', 'apr-cc-pV(Q+d)Z', 'aug-cc-pCV5Z', 'aug-cc-pCVDZ', 'aug-cc-pCVQZ', 'aug-cc-pCV(T+d)Z', 'aug-cc-pCVTZ', 'aug-cc-pV(5+d)Z', 'aug-cc-pV5Z', 'aug-cc-pV(6+d)Z', 'aug-cc-pV6Z', 'aug-cc-pV(D+d)Z', 'aug-cc-pVDZ', 'aug-cc-pV(Q+d)Z', 'aug-cc-pVQZ', 'aug-cc-pV(T+d)Z', 'aug-cc-pVTZ', 'aug-cc-pVTZ-J', 'aug-cc-pwCV5Z', 'aug-cc-pwCV5Z-NR', 'aug-cc-pwCVDZ', 'aug-cc-pwCVQZ', 'aug-cc-pwCVQZ-NR', 'aug-cc-pwCVTZ', 'aug-cc-pwCVTZ-NR', 'aug-mcc-pV5Z', 'aug-mcc-pV6Z', 'aug-mcc-pV7Z', 'aug-mcc-pV8Z', 'aug-mcc-pVQZ', 'aug-mcc-pVTZ', 'aug-pc-0', 'aug-pc-1', 'aug-pc-2', 'aug-pc-3', 'aug-pc-4', 'aug-pcJ-0', 'aug-pcJ-0_2006', 'aug-pcJ-1', 'aug-pcJ-1_2006', 'aug-pcJ-2', 'aug-pcJ-2_2006', 'aug-pcJ-3', 'aug-pcJ-3_2006', 'aug-pcJ-4', 'aug-pcJ-4_2006', 'aug-pcS-0', 'aug-pcS-1', 'aug-pcS-2', 'aug-pcS-3', 'aug-pcS-4', 'aug-pV7Z', 'B2 basis set for Zn', 'Bauschlicher ANO', 'Binning/Curtiss SV', 'Binning/Curtiss SVP', 'Binning/Curtiss VTZ', 'Binning/Curtiss VTZP', 'cc-pCV5Z', 'cc-pCV6Z', 'cc-pCV6Z(old)', 'cc-pCVDZ', 'cc-pCVDZ(old)', 'cc-pCVQZ', 'cc-pCVQZ(old)', 'cc-pCVTZ', 'cc-pCVTZ(old)', 'cc-pV(5+d)Z', 'cc-pV5Z', 'cc-pV(6+d)Z', 'cc-pV6Z', 'cc-pV8Z', 'cc-pV9Z', 'cc-pV(D+d)Z', 'cc-pVDZ', 'cc-pVDZ(seg-opt)', 'cc-pV(Q+d)Z', 'cc-pVQZ', 'cc-pVQZ(seg-opt)', 'cc-pV(T+d)Z', 'cc-pVTZ', 'cc-pVTZ(seg-opt)', 'cc-pwCV5Z', 'cc-pwCV5Z Core Set', 'cc-pwCV5Z-NR', 'cc-pwCVDZ', 'cc-pwCVQZ', 'cc-pwCVQZ-NR', 'cc-pwCVTZ', 'cc-pwCVTZ-NR', 'ccemd-2', 'ccemd-3', 'ccJ-pV5Z', 'ccJ-pVDZ', 'ccJ-pVQZ', 'ccJ-pVTZ', 'coemd-2', 'coemd-3', 'coemd-4', 'coemd-ref', 'CVTZ', 'd-aug-cc-pV5Z', 'd-aug-cc-pV6Z', 'd-aug-cc-pVDZ', 'd-aug-cc-pVQZ', 'd-aug-cc-pVTZ', 'Def2-QZVP', 'Def2-QZVPD', 'Def2-SVP', 'def2-SV(P)', 'Def2-SVPD', 'Def2-TZVP', 'Def2-TZVPD', 'dhf-QZVP', 'dhf-SV(P)', 'dhf-TZVP', 'Dunning-Hay Double Rydberg', 'Dunning-Hay Rydberg', 'DZ + Double Rydberg (Dunning-Hay)', 'DZ + Rydberg (Dunning)', 'DZ (Dunning)', 'DZP + Rydberg (Dunning)', 'DZP (Dunning)', 'DZQ', 'DZVP2 (DFT Orbital)', 'DZVP (DFT Orbital)', 'Feller Misc. CVDZ', 'Feller Misc. CVQZ', 'Feller Misc. CVTZ', 'GAMESS PVTZ', 'GAMESS VTZ', 'IGLO-II', 'IGLO-III', 'jul-cc-pV(D+d)Z', 'jul-cc-pV(Q+d)Z', 'jul-cc-pV(T+d)Z', 'jun-cc-pV(D+d)Z', 'jun-cc-pV(Q+d)Z', 'jun-cc-pV(T+d)Z', 'LANL08', 'LANL08+', 'LANL08d', 'LANL08(f)', 'Lanl2-\[10s8p7d3f2g\]', 'Lanl2-\[5s4p4d2f\]', 'Lanl2-\[6s4p4d2f\]', 'Lanl2DZ+1d1f', 'Lanl2DZ+2s2p2d2f', 'LANL2TZ', 'LANL2TZ+', 'LANL2TZ(f)', 'm6-31G', 'm6-31G*', 'maug-cc-pV(D+d)Z', 'maug-cc-pVDZ', 'maug-cc-pV(Q+d)Z', 'maug-cc-pVQZ', 'maug-cc-pV(T+d)Z', 'maug-cc-pVTZ', 'may-cc-pV(Q+d)Z', 'may-cc-pV(T+d)Z', 'McLean/Chandler VTZ', 'MG3S', 'MIDI!', 'MIDI (Huzinaga)', 'MINI (Huzinaga)', 'MINI (Scaled)', 'modified LANL2DZ', 'NASA Ames ANO', 'NASA Ames ANO2', 'NASA Ames cc-pCV5Z', 'NASA Ames cc-pCVQZ', 'NASA Ames cc-pCVTZ', 'NASA Ames cc-pV5Z', 'NASA Ames cc-pVQZ', 'NASA Ames cc-pVTZ', 'Partridge Uncontracted 1', 'Partridge Uncontracted 2', 'Partridge Uncontracted 3', 'Partridge Uncontracted 4', 'pc-0', 'pc-1', 'pc-2', 'pc-3', 'pc-4', 'pcemd-2', 'pcemd-3', 'pcemd-4', 'pcJ-0', 'pcJ-0_2006', 'pcJ-1', 'pcJ-1_2006', 'pcJ-2', 'pcJ-2_2006', 'pcJ-3', 'pcJ-3_2006', 'pcJ-4', 'pcJ-4_2006', 'pcS-0', 'pcS-1', 'pcS-2', 'pcS-3', 'pcS-4', 'pSBKJC', 'Pt - mDZP', 'pV6Z', 'pV7Z', 'Roos_ANO_DZ', 'Roos_ANO_TZ', 'Roos Augmented Double Zeta ANO', 'Roos Augmented Triple Zeta ANO', 's3-21G', 's3-21G*', 's6-31G', 's6-31G*', 'Sadlej pVTZ', 'SDB-aug-cc-pVQZ', 'SDB-aug-cc-pVTZ', 'SDB-cc-pVQZ', 'SDB-cc-pVTZ', 'STO-2G', 'STO-3G', 'STO-3G*', 'STO-6G', 'SV + Double Rydberg (Dunning-Hay)', 'SV + Rydberg (Dunning-Hay)', 'SV (Dunning-Hay)', 'SVP + Rydberg (Dunning-Hay)', 'SVP (Dunning-Hay)', 'TZ (Dunning)', 'TZVP (DFT Orbital)', 'UGBS', 'un-ccemd-ref', 'un-pcemd-ref', 'Wachters+f', 'WTBS', 'Z3Pol' ]

        if self.basis_set.lower() not in [x.lower() for x in available_basis_sets]:
            print("basis set ", self.basis_set, " not available / not supported")
            return None
        if self.func.lower() not in [x.lower() for x in available_functionals]:
            print("density functional ", self.func, " not available / not supported")
            return None
        return True


    def reset(self):
        NWChem.reset(self)
        self.atoms = None
        self.results = {}
        self.initialized = False
        self.orbitals = []


    def initialize(self, atoms=None):
        return


    def write_input(self, atoms, properties=None, system_changes=None):
        NWChem.write_input(self, atoms, properties=properties, system_changes=system_changes)
        filename = self.title+'.nw'
        buf = []
        drvset = False
        for line in open(filename,'r'):
            buf.append(line)
            if re.search('^dft\s*$',line):
                buf.append('  print \"final vectors analysis\" \"final vectors\"\n')
                buf.append('  direct\n')
                buf.append('  noio\n')
                buf.append('  iterations ' + str(self.scfiterlimit) + '\n')
            if re.search('^end\s*$',line) and self.task == 'optimize' and not drvset:
                buf.append('\ndriver\n')
                if self.force_conv > 0.000450:
                    buf.append('  loose\n')
                if self.maxiter:
                    buf.append('  maxiter ' + str(self.maxiter) + '\n')
                buf.append('end\n\n')
                drvset = True

        outfile = open(filename,'w')
        outfile.write("".join(buf))
        outfile.close()


    def get_potential_energy(self,atoms):
        if self.task == 'optimize':
            energy = NWChem.get_potential_energy(self,atoms)
            self.read_geometry()
            atoms.set_positions(self.atoms.get_positions())
            return energy
        else:
            return NWChem.get_potential_energy(self,atoms)


    def get_mo(self,mo_type,atoms):
        if not self.orbitals:
            self.energy = atoms.get_potential_energy()
        assert mo_type in ['HOMO', 'LUMO']
        if mo_type == 'HOMO':
            orb_energy = None
            orb_number = None
            for orbital in self.orbitals:
                if orbital['occupation'] > 0:
                    if orbital['energy'] > orb_energy:
                        orb_energy = orbital['energy']
                        orb_number = orbital['number']
        if mo_type == 'LUMO':
            orb_energy = None
            orb_number = None
            for orbital in self.orbitals:
                if orbital['occupation'] < 1:
                    if orbital['energy'] < orb_energy or not orb_energy:
                        orb_energy = orbital['energy']
                        orb_number = orbital['number']
        return [orb_number, orb_energy*Hartree]


    def get_mos(self, atoms):
        if not self.orbitals:
            self.energy = atoms.get_potential_energy()
        orbital_coefficients = []
        eigen_energies = []
        for orbital in self.orbitals:
            orbital_coefficients.append(orbital['coefficients'])
            eigen_energies.append(orbital['energy'])
        return [np.array(orbital_coefficients), np.diag(np.array(eigen_energies)*Hartree)]


    def read_results(self):
        NWChem.read_results(self)
        self.read_mos()


    def read_energy(self): # reads the last energy from geometry optimization output
        NWChem.read_energy(self)
        text = open(self.label + '.out', 'r').read()
        lines = iter(text.split('\n'))

        # Energy:
        estring = 'Total '
        if self.parameters.xc == 'RHF':
            estring += 'SCF'
        elif self.parameters.xc == 'MP2':
            estring += 'MP2'
        else:
            estring += 'DFT'
        estring += ' energy'
        for line in lines:
            if line.find(estring) >= 0:
                energy = float(line.split()[-1])
                continue
        self.results['energy'] = energy*Hartree


    def read_geometry(self):
        text = open(self.label + '.out', 'r').read()
        lines = iter(text.split('\n'))
        natoms = len(self.atoms)
        read_flag = False
        atoms_read = 0
        xyz_string = str(natoms)+'\n\n'
        for line in lines:
            if 'Output coordinates in angstroms' in line:
                read_flag = True
                xyz_string = str(natoms)+'\n\n'
                continue
            if read_flag:
                match = re.search('(\d+)\s+(\w+)\s+([+-]?\d+\.\d+)\s+([+-]?\d+\.\d+)\s+([+-]?\d+\.\d+)\s+([+-]?\d+\.\d+)',line) 
                if match:
                    string = match.group(2) + ' ' + match.group(4) + ' ' + match.group(5) + ' ' + match.group(6) + '\n'
                    xyz_string += string
                    atoms_read += 1
                    assert int(match.group(1)) == natoms, 'number of xyz lines more than expected number of atoms'
            if atoms_read == natoms:
                read_flag = False
                atoms_read = 0
        new_atoms = read(StringIO(xyz_string), format='xyz')
        self.atoms.set_positions(new_atoms.get_positions())

# end of class KITnwchem



import json
import numpy as np
from ase import Atoms

""" This module includes generic data input/output/manipulation methods. """

def translate_keys(dct, key, transl):
    """ Translates all keys which match 'key' in 'dct' into 'transl' """
    new = {}
    for k, val in dct.items():
        if isinstance(val, dict):
            newval = translate_keys(val, key, transl)
        elif isinstance(val, list):
            newval = []
            for item in val:
                if isinstance(item, dict):
                    newval.append(translate_keys(item, key, transl))
                else:
                    newval.append(item)
        else:
            newval = val
        if k == key:
            new[transl] = newval
        else:
            new[k] = newval
    return new

def ddump(obj):
    """ Dumps a dictionary of objects attributes """
    dct = {}
    lst = [
        attr for attr in dir(obj) if not attr.startswith('__') 
        and not callable(getattr(obj,attr))
        ]
    for item in lst:
        o = getattr(obj,item)
        if type(o) in [bool, int, float, str, None]:
            dct[item] = o
        if type(o) in [list, tuple, dict]:
            dct[item] = o
        elif hasattr(o, 'todict'):
            dct[item] = o.todict()
        elif o.__class__ == np.ndarray:
            dct[item] = o.tolist()
        else:
            pass
    return dct

def jdefault(obj):
    """ Returns a JSON serializable object """
    from ase import Atoms
    from ase.db.row import atoms2dict

    if isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, Atoms):
        return atoms2dict(obj)
    elif hasattr(obj, 'todict'):
        return obj.todict()
    elif hasattr(obj, '__dict__'):
        return obj.__dict__
    else:
        raise TypeError('Not serializable')
'''
def json_load_byteified(file_handle):
    return _byteify(
        json.load(file_handle, object_hook=_byteify),
        ignore_dicts=True
    )

def json_loads_byteified(json_text):
    return _byteify(
        json.loads(json_text, object_hook=_byteify),
        ignore_dicts=True
    )


def _byteify(data, ignore_dicts = False):
    if isinstance(data, unicode):
        return data.encode('utf-8')
    if isinstance(data, list):
        return [ _byteify(item, ignore_dicts=True) for item in data ]
    if isinstance(data, dict) and not ignore_dicts:
        return {
            _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
            for key, value in data.items()
        }
    return data
'''

def single_entry(entry):
    if isinstance(entry, list):
        assert len(entry) == 1, 'can process only one entry at a time'
        entry = entry[0]
    return entry

def dict_update(dct, maplist, filename):
    """ In-place update of dictionary datadict at position specified by 
    maplist with data from file, i.e. dct[maplist[0]][maplist[1]]... = data
    """
    from functools import reduce
    import operator

    with open(filename, 'r') as inp:
        data = json.load(inp)

    if isinstance(data, dict):
        reduce(operator.getitem, maplist[:-1], dct)[maplist[-1]].update(data)
    else:
        reduce(operator.getitem, maplist[:-1], dct)[maplist[-1]] = data

class GeneratorCounter(object):
    """ wrap a generator object to count the number of elements yielded """
    def __init__(self, generator):
        self.total = 0
        self.generator = generator

    def __iter__(self):
        return self

    def __next__(self):
        nxt = next(self.generator)
        self.total += 1
        return nxt

    def __len__(self):
        """ return the current number of elements yielded """
        return self.total

def cleanup_atoms_dict(adict):
    """ delete keywords in dictionary not allowed in Atoms.__init__ """

    dellist = ['unique_id', 'magmom', 'calculator', 'calculator_parameters',
               'energy', 'forces', 'dipole', 'stress']
    subdict = {'initial_magmoms': 'magmoms'}
    for key in subdict:
        if key in adict:
            adict[subdict[key]] = adict.pop(key)
    for key in dellist:
        if key in adict:
            del adict[key]
    return adict

@classmethod
def from_dict(cls, adict):
    """ deserialize json serializable dictionaries to atoms objects """
    from ase.io import jsonio    
    from ase.constraints import dict2constraint

    adict_ = jsonio.decode(jsonio.encode(cleanup_atoms_dict(adict)))
    if 'constraints' in adict_:
        constr = adict_.pop('constraints')
        if isinstance(constr, list):
            adict_['constraint'] = [dict2constraint(c) for c in constr]
        else:
            adict_['constraint'] = [dict2constraint(constr)]

    return cls(**adict_)

Atoms.from_dict = from_dict

def atoms2dict(atoms):
    """ convert Atoms object to a dictionary used to recreate the object """
    from ase.io import jsonio
    from ase.db.row import atoms2dict
    import json

    adict = cleanup_atoms_dict(atoms2dict(atoms))
    return json.loads(jsonio.encode(adict))



"""A simple workflow engine"""

import threading
import uuid
import os
import subprocess as sp
from time import sleep
from fireworks.core.rocket_launcher import launch_rocket
from fireworks.queue.queue_launcher import launch_rocket_to_queue
from fireworks import Workflow, Firework, PyTask, FWorker
from fireworks.core.launchpad import LaunchPad
from fireworks.utilities.fw_serializers import FWSerializable
from fireworks.utilities.fw_serializers import recursive_serialize
from fireworks.utilities.fw_serializers import recursive_deserialize
from fireworks.utilities.fw_utilities import get_fw_logger
from fireworks.utilities.fw_utilities import create_datestamp_dir
from prettytable import PrettyTable
from wfgenes.query.wfquery import WFQuery




class WFEngine(FWSerializable):
    """A simple engine to manage workflows"""

    thread = None
    event = None

    def __init__(self, launchpad, qadapter=None, wf_query=None, name=None,
                 launchdir='.', sleep_time=30):
        """
        Args:
            launchpad (LaunchPad): a launchpad to establish connection to a
                FireWorks database
            qadapter (CommonAdapter, None): a qadapter for submitting batch jobs
            wf_query (dict, None): a workflow query
            name (str, None): name of the engine; if None one will be generated
            launchdir (str): launch directory for both interactive and batch jobs
            sleep_time (int): launcher thread awakes every `sleep_time` seconds
        """
        self.launchpad = launchpad
        self.qadapter = qadapter
        self.wf_ids = ([] if wf_query is None else
                       self.launchpad.get_wf_ids(wf_query))
        self.launchdir = launchdir
        self.sleep_time = sleep_time
        self.name = name if name is not None else str(uuid.uuid4())
        if qadapter is not None:
            self.fworker_qlaunch = FWorker(name=self.name, category='batch')
        else:
            self.fworker_qlaunch = None
        self.fworker_rlaunch = FWorker(name=self.name, category='interactive')
    @property
    def name(self):
        """get the name of the engine"""
        return self.__name

    @name.setter
    def name(self, new_name):
        """set the name of the engine"""
        self.__name = new_name
        self.fworker_qlaunch = FWorker(name=self.__name, category='batch')
        self.fworker_rlaunch = FWorker(name=self.__name, category='interactive')

    def show_nodes_status(self):
        """Display the status summary of the nodes"""
        wfq = WFQuery(self.launchpad, wf_query={'nodes': {'$in': self.wf_ids}})
        fw_info = wfq.get_fw_info()
        if len(fw_info) > 0:
            columns = fw_info[0].keys()
            table = PrettyTable(columns)
            for fwk in fw_info:
                table.add_row([fwk[i] for i in columns])
            print(table)
        else:
            print('No nodes')

    def show_wf_status(self):
        """Display the status summary of the workflows"""
        wfq = WFQuery(self.launchpad, wf_query={'nodes': {'$in': self.wf_ids}})
        wfq.get_wf_info()

    def status_summary(self):
        """Display a status summary of workflows and nodes"""
        self.show_wf_status()
        print('Nodes summary:')
        self.show_nodes_status()

    def show_launcher_status(self):
        """Check whether a launcher thread is running"""
        if self.thread:
            if self.thread.is_alive():
                print('launcher thread is currently running')
            else:
                print('launcher thread not started')
        else:
            print('launcher thread not created')

    def status_detail(self, *fw_ids):
        """
        Print a detailed status of specified nodes

        Args:
            fw_ids ([int]): One or more fw_ids of the nodes
        """
        for fw_id in fw_ids:
            fw_dict = self.launchpad.get_fw_dict_by_id(fw_id)
            if fw_dict['state'] == 'FIZZLED':
                launch = fw_dict['launches'][-1]
                exception = launch['action']['stored_data']['_exception']
                print(str(fw_id), exception['_stacktrace'])
                
            return fw_dict    

    def get_failed(self):
        """
        Get failed job ids

        Returns:
            ([int]): a list of fw_ids of failed jobs
        """
        wfq = WFQuery(self.launchpad, wf_query={'nodes': {'$in': self.wf_ids}},
                      fw_query={'state': 'FIZZLED'})
        return wfq.get_fw_ids()

    def qlaunch(self, fw_id):
        """
        Launch a batch node by submitting a job to the queuing system

        Args:
            fw_id (int): a fwd_id of the node to launch
        """
        assert fw_id in self.get_fw_ids(), 'invalid fw_id'
        assert self.qadapter is not None, 'qadapter is not defined'
        assert self.fworker_qlaunch is not None, 'qlaunch worker is not defined'
        launch_rocket_to_queue(self.launchpad, self.fworker_qlaunch,
                               self.qadapter, reserve=True,
                               create_launcher_dir=True,
                               launcher_dir=self.launchdir,
                               fw_id=fw_id)

    def rlaunch(self, fw_id):
        """
        Launch an interactive node

        Args:
            fw_id (int): a fwd_id of the node to launch
        """
        assert fw_id in self.get_fw_ids(), 'invalid fw_id'
        init_dir = os.getcwd()
        log_dir = self.launchpad.get_logdir()
        logger = get_fw_logger('interactive', l_dir=log_dir)
        launch_dir = create_datestamp_dir(init_dir, logger, prefix='launcher_')
        try:
            os.chdir(launch_dir)
            launch_rocket(self.launchpad, self.fworker_rlaunch, fw_id)
        except Exception:
            os.chdir(init_dir)
            raise
        finally:
            os.chdir(init_dir)

    def check_lostjobs(self, time=14400):
        """
        Detect nodes that have been launched but not updated within the
        specified time. The state of such nodes is set to FIZZLED.

        Args:
            time (int): minimim time in seconds since last update, default: 4h

        Returns:
            lost_fw_ids ([int]): a list of fw_ids of the lost runs
        """
        fw_query = {'fw_id': {'$in': self.get_fw_ids()}}
        lostjobs = self.launchpad.detect_lostruns(expiration_secs=time,
                                                  fizzle=True, query=fw_query)
        lost_fw_ids = lostjobs[1]
        if len(lost_fw_ids) != 0:
            string = 'Lost jobs detected: {}.'
            print(string.format(lost_fw_ids))  # this should be made as log
        return lost_fw_ids

    @recursive_serialize
    def to_dict(self):
        """
        Dump the engine object to a dictionary

        Returns:
            (dict): a dictionary with all parameters needed to call __init__
        """
        return {'launchpad': self.launchpad,
                'qadapter': self.qadapter,
                'wf_query': {'nodes': {'$in': self.wf_ids}},
                'launchdir': self.launchdir,
                'name': self.__name,
                'sleep_time': self.sleep_time}

    @classmethod
    @recursive_deserialize
    def from_dict(cls, m_dict):
        """
        Construct an engine from a dictionary

        Args:
            m_dict (dict): a dict with all parameters needed to call __init__

        Returns:
            FWEngine object
        """
        launchpad = LaunchPad.from_dict(m_dict.get('launchpad'))
        qadapter = m_dict.get('qadapter')
        wf_query = m_dict.get('wf_query')
        name = m_dict.get('name')
        launchdir = m_dict.get('launchdir')
        sleep_time = m_dict.get('sleep_time')
        return cls(launchpad=launchpad, qadapter=qadapter, wf_query=wf_query,
                   name=name, launchdir=launchdir, sleep_time=sleep_time)

    def start(self):
        """Start a launcher thread"""
        if self.thread and self.thread.is_alive():
            print('Warning: launcher thread is already running')
        else:
            self.event = threading.Event()
            self.thread = threading.Thread(target=self.launcher,
                                           args=(self.event, ))
            self.thread.start()

            
    def stop(self):
        """Gracefully stop the launcher thread if it is running"""
        if self.event and not self.event.is_set():
            self.event.set()
            print('stopping the launcher thread')

    def launcher(self, stop_event):
        """
        The main loop of the launcher

        Args:
            stop_event (threading.Event): an object used to quit the launcher
        """
        while not stop_event.is_set():
            wf_query = {'nodes': {'$in': self.wf_ids}}
            fw_query = {'state': 'READY',
                        'spec._category': {'$in': ['batch', 'interactive']}}
            fw_ids = self.launchpad.get_fw_ids_in_wfs(wf_query, fw_query)
            for fw_id in fw_ids:
                input_config = self.launchpad.get_fw_dict_by_id(fw_id)
                category = input_config['spec'].get('_category', None)
                if category == 'batch':
                    # Launch the job on the cluster
                    self.qlaunch(fw_id)
                else:
                    # Launch the job locally
                    self.rlaunch(fw_id)
            sleep(self.sleep_time)
        print('launcher thread stopped')

    def get_fw_ids(self):
        """
        Return the list of nodes currently included in the engine

        Returns:
            (list): fw_ids of the nodes
        """
        wf_query = {'nodes': {'$in': self.wf_ids}}
        return self.launchpad.get_fw_ids_in_wfs(wf_query)

    def add_node(self, func, inputs, outputs=[], name=None, kwargs={}):
        """
        Add a python function node to an existing workflow

        Args:
            func (str): a function name with an optional module name in the
                format 'module.function'
            inputs ([tuple]): a list of positional arguments for the provided
                function. Every input is described by a tuple
                (fw_id, name, value) with the following elements:
                    The fw_id of a parent node providing the input; if the
                        input is provided as a constant value, then None should
                        be specified.
                    The name of the input as provided in the list of outputs of
                    the parent node;
                    The value of the input; if output data from a parent node
                        is used as input, then this should be set to None.

            outputs ([str]): names of the outputs
        """

        node_ids = list(set(i[0] for i in inputs if i[0] is not None))
        # check that all parent nodes are in the managed wfs
        msg = 'some parent nodes are not in engine'
        assert all(n in self.get_fw_ids() for n in node_ids), msg
        # check that all parent nodes are in one worfklow
        qres = self.launchpad.get_wf_ids({'nodes': {'$in': node_ids}})
        assert len(qres) > 0, 'no valid parent nodes defined'
        assert len(qres) == 1, 'some parent nodes are not in one workflow'
        # later, add all upstream fireworks of the nodes from other workflows
        # later check the outputs of parent nodes to match the inputs
        inps = [i[1] for i in inputs]
        # check that any equal inputs have the same source
        msg = 'input "{}" has more than one source'
        for inp in set(inps):
            sources = set(i[0] for i in inputs if i[1] == inp)
            assert len(sources) == 1, msg.format(inp)

        task = PyTask(func=func, inputs=inps, outputs=outputs, kwargs=kwargs)
        spec = {}
        spec['_dupefinder'] = {'_fw_name': 'DupeFinderExact'}
        spec['_category'] = 'interactive'  # fix later as arg
        # spec['_fworker'] = ? (if category not interactive) fix this
        # spec['_qadapter'] = ? (if category not interactive) fix this
        locs = {i[1]: i[2] for i in inputs if i[0] is None}
        spec.update(locs)
        wflow = Workflow([Firework(tasks=[task], spec=spec, name=name)])
        self.launchpad.append_wf(wflow, fw_ids=node_ids)

    def add_workflow(self, workflow=None, fw_id=None):
        """
        Add a workflow to the engine
        Either a workflow object or a fw_id must be defined.

        Args:
            workflow (Workflow, None): a workflow object
            fw_id (int, None): a fw_id of a workflow existing on the launchpad
        """
        assert workflow is not None or fw_id is not None
        assert not (workflow is not None and fw_id is not None)
        if fw_id is None:
            assert isinstance(workflow, (Workflow, Firework))
            fw_id = list(self.launchpad.add_wf(workflow).values())[0]
        else:
            assert isinstance(fw_id, int)
            assert fw_id not in self.get_fw_ids(), 'workflow already in engine'
            fw_ids = self.launchpad.get_fw_ids({'fw_id': fw_id})
            assert len(fw_ids) == 1, 'no workflow with this id'
        self.wf_ids.append(fw_id)
        self.launchpad.m_logger.info('Added wf_id: {}'.format(fw_id))

    def remove_workflow(self, fw_id):
        """
        Remove a workflow from the engine (but not deleted from launchpad)

        Args:
            fw_id (int): a fw_id of a node in the workflow to remove
        """
        self.wf_ids.remove(fw_id)
        self.launchpad.m_logger.info('Removed wf_id: {}'.format(fw_id))

    def update_node(self, fw_id, update_dict):
        """
        Update (modify) a workflow node
        Only nodes in WAITING, READY and FIZZLED states can be modified.

        Args:
           fw_id (int): the fw_id of the node to modify
           update_dict (dict): a dictionary with the updates to perform
        """
        state = self.launchpad.get_fw_by_id(fw_id).state
        assert state in ['WAITING', 'READY', 'FIZZLED']
        self.launchpad.update_spec([fw_id], update_dict)

    def rerun_node(self, fw_id):
        """
        Rerun a workflow node
        Only nodes in COMPLETED and FIZZLED states can be rerun.

        Args:
            fw_id: the fw_id of the node to rerun
        """
        state = self.launchpad.get_fw_by_id(fw_id).state
        assert state in ['COMPLETED', 'FIZZLED', 'RESERVED']
        self.launchpad.rerun_fw(fw_id)

    def update_rerun_node(self, fw_id, update_dict):
        """
        Update (modify) and rerun a workflow node combined in one function
        Only nodes in COMPLETED, WAITING, READY and FIZZLED states can be
        processed.

        Args:
            fw_id (int): the fw_id of the node to process
            update_dict (dict): a dictionary with the updates to perform
        """
        state = self.launchpad.get_fw_by_id(fw_id).state
        assert state in ['COMPLETED', 'WAITING', 'READY', 'FIZZLED']
        if state == 'COMPLETED':
            self.launchpad.defuse_fw(fw_id)
            self.launchpad.update_spec([fw_id], update_dict)
            self.launchpad.reignite_fw(fw_id)
        elif state in ['WAITING', 'READY']:
            self.launchpad.update_spec([fw_id], update_dict)
        elif state == 'FIZZLED':
            self.launchpad.update_spec([fw_id], update_dict)
            self.launchpad.rerun_fw(fw_id)

    def cancel_job(self, fw_id, restart=False, pause=False):
        """
        Cancel the execution of a node in RESERVED or RUNNING state
        Either restart or pause can be set to True if required.

        Args:
            fw_id (int): the fw_id of the node to cancel
            restart (bool): rerun after cancelling a RUNNING node
            pause (bool): pause after cancelling a RUNNING node
        """
        assert not (restart and pause)
        state = self.launchpad.get_fw_by_id(fw_id).state
        assert state in ['RESERVED', 'RUNNING'], 'Job not in necessary state'
        fw_dict = self.launchpad.get_fw_dict_by_id(fw_id)
        category = fw_dict['spec'].get('_category', None)
        assert category == 'batch', 'Not a batch job'
        reserve_id = self.launchpad.get_reservation_id_from_fw_id(fw_id)
        self.exec_cancel(reserve_id)
        self.check_jobcancel(reserve_id)
        if state == 'RESERVED':
            self.launchpad.cancel_reservation_by_reservation_id(reserve_id)
        elif state == 'RUNNING':
            assert (restart is True or pause is True), 'Either restart or pause should be true'
            if restart:
                self.launchpad.rerun_fw(fw_id)
            else:
                self.launchpad.defuse_fw(fw_id)

    def exec_cancel(self, res_id):
        """Execute the slurm cancel command"""
        if os.system(f'scancel {res_id}') != 0:
            raise RuntimeError('Error executing command \"scancel\"')
        self.launchpad.m_logger.info('Cancelled res_id: {}'.format(res_id))

    def check_jobcancel(self, res_id):
        """Execute the slurm sacct command"""
        output = sp.getoutput(f'sacct -j {res_id}')
        if 'CANCELLED' not in output:
            raise RuntimeError('Error job is not cancelled')
    

